import json
import os
from copy import deepcopy
from typing import Any, Dict, Sequence, Set, Union

from funcy import omit
from pydantic import BaseModel


def show_difference(left: Dict, right: Dict):
    left_ = deepcopy(left)
    right_ = deepcopy(right)
    for key in list(left_.keys()):
        if left_.get(key) == right_.get(key):
            left_.pop(key)
            right_.pop(key)
    return f"\n{left_} \n!=\n {right_}"


def assert_equals(left, right):
    # separate function is helpful for debug
    # cause you see dicts without skip_keys
    assert left == right, show_difference(left, right)


def check_obj(
    obj: Union[BaseModel, Dict[str, Any]],
    values: Dict[str, Any],
    skip_keys: Union[Set[str], Sequence[str]],
):
    if isinstance(obj, BaseModel):
        obj_values = obj.dict(exclude=set(skip_keys))
    else:
        obj_values = omit(obj, skip_keys)
    values = omit(values, skip_keys)
    assert_equals(obj_values, values)


def convert_objects_to_str_in_json_serializable_object(
    o: Union[list, dict]
) -> Union[list, dict]:
    return json.loads(json.dumps(o, default=str))


def is_os_windows_and_py_lt_3_9() -> bool:
    return (
        os.environ.get("GITHUB_MATRIX_OS") == "windows-latest"
        and os.environ.get("GITHUB_MATRIX_PYTHON", "") < "3.9"
    )


def is_os_windows_and_py_lt_3_8() -> bool:
    return (
        os.environ.get("GITHUB_MATRIX_OS") == "windows-latest"
        and os.environ.get("GITHUB_MATRIX_PYTHON", "") < "3.8"
    )



import os
import sys
from time import sleep
from typing import Callable, Tuple

import git
import pytest
from click.testing import Result
from typer.testing import CliRunner

import gto
from gto.cli import app
from gto.config import CONFIG_FILE_NAME


class Runner:
    def __init__(self):
        self._runner = CliRunner(mix_stderr=False)

    def invoke(self, *args, **kwargs) -> Result:
        return self._runner.invoke(app, *args, **kwargs)


@pytest.fixture
def runner() -> Runner:
    return Runner()


@pytest.fixture(autouse=True)
def set_recursion_limit() -> None:
    sys.setrecursionlimit(2000)


@pytest.fixture
def empty_git_repo(tmp_path):
    repo = git.Repo.init(tmp_path)

    def write_file(name, content):
        fullpath = os.path.join(tmp_path, name)
        os.makedirs(os.path.dirname(fullpath), exist_ok=True)
        with open(fullpath, "w", encoding="utf8") as file:
            file.write(content)

    return repo, write_file


@pytest.fixture
def repo_with_commit(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo

    write_file(
        "some-random-file",
        "some-random-text",
    )
    repo.index.add(["some-random-file"])
    repo.index.commit("Initial commit")

    return repo, write_file


@pytest.fixture
def init_showcase_semver(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo

    write_file(
        CONFIG_FILE_NAME,
        "",
    )

    return repo, write_file


@pytest.fixture
def showcase(
    init_showcase_semver,
):  # pylint: disable=too-many-locals, too-many-statements
    repo: git.Repo
    repo, write_file = init_showcase_semver
    path = repo.working_dir

    write_file("models/random-forest.pkl", "1st version")
    write_file("models/neural-network.pkl", "1st version")
    repo.index.add(["models"])
    repo.index.commit("Create models")

    gto.api.annotate(
        path, "rf", type="model", path="models/random-forest.pkl", must_exist=True
    )
    gto.api.annotate(
        path, "nn", type="model", path="models/neural-network.pkl", must_exist=True
    )
    gto.api.annotate(path, "features", type="dataset", path="datasets/features.csv")

    repo.index.add(["artifacts.yaml"])
    first_commit = repo.index.commit("Add artifacts")

    nn_vname = "v0.0.1"
    rf_vname = "v1.2.3"
    gto.api.register(path, "rf", "HEAD", rf_vname)
    gto.api.register(path, "nn", "HEAD")
    sleep(1)

    write_file("models/random-forest.pkl", "2nd version")

    second_commit = repo.index.commit("Update model")

    # bump version automatically
    gto.api.register(path, "rf", "HEAD")

    gto.api.assign(path, "nn", "staging", version=nn_vname)
    gto.api.assign(path, "rf", "production", version=rf_vname)
    sleep(1)  # this is needed to ensure right order of assignments in later checks
    # the problem is git tags doesn't have miliseconds precision, so we need to wait a bit
    gto.api.assign(path, "rf", "staging", ref="HEAD")
    sleep(1)
    gto.api.assign(path, "rf", "production", ref=repo.head.ref.commit.hexsha)
    sleep(1)
    gto.api.assign(path, "rf", "production", version=rf_vname, force=True)
    return path, repo, write_file, first_commit, second_commit



import pytest

from tests.utils import (
    is_os_windows_and_py_lt_3_8,
    is_os_windows_and_py_lt_3_9,
)

skip_for_windows_py_lt_3_9 = pytest.mark.skipif(
    is_os_windows_and_py_lt_3_9(),
    reason="functionality requires python features not working on windows with python<3.9",
)

only_for_windows_py_lt_3_8 = pytest.mark.skipif(
    not is_os_windows_and_py_lt_3_8(),
    reason="test makes sense only for windows with python<3.9",
)



import logging
from collections import defaultdict
from functools import partial, wraps
from gettext import gettext
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import click
import typer
from click import Abort, ClickException, Command, Context, HelpFormatter
from click.exceptions import Exit
from typer import Argument, Option, Typer
from typer.core import TyperCommand, TyperGroup

import gto
from gto.constants import (
    ASSIGNMENTS_PER_VERSION,
    VERSIONS_PER_STAGE,
    VersionSort,
)
from gto.exceptions import GTOException, NotImplementedInGTO, WrongArgs
from gto.ui import (
    EMOJI_FAIL,
    EMOJI_GTO,
    EMOJI_OK,
    bold,
    cli_echo,
    color,
    echo,
    stderr_echo,
)
from gto.utils import format_echo, make_ready_to_serialize


class CommandGroups:
    querying = "Read the registry"
    modifying = "Modify artifacts"
    enriching = "Manage artifact enrichments"


class GtoFormatter(click.HelpFormatter):
    def write_heading(self, heading: str) -> None:
        super().write_heading(bold(heading))


class GtoCliMixin(Command):
    def __init__(
        self,
        name: Optional[str],
        examples: Optional[str],
        section: str = "other",
        aliases: List[str] = None,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self.examples = examples
        self.section = section
        self.aliases = aliases

    # def collect_usage_pieces(self, ctx: Context) -> List[str]:
    #     return [p.lower() for p in super().collect_usage_pieces(ctx)]

    def get_help(self, ctx: Context) -> str:
        """Formats the help into a string and returns it.

        Calls :meth:`format_help` internally.
        """
        formatter = GtoFormatter(
            width=ctx.terminal_width, max_width=ctx.max_content_width
        )
        self.format_help(ctx, formatter)
        return formatter.getvalue().rstrip("\n")

    def format_epilog(self, ctx: Context, formatter: HelpFormatter) -> None:
        super().format_epilog(ctx, formatter)
        if self.examples:
            with formatter.section("Examples"):
                formatter.write(self.examples)


def _extract_examples(
    help_str: Optional[str],
) -> Tuple[Optional[str], Optional[str]]:
    if help_str is None:
        return None, None
    try:
        examples = help_str.index("Examples:")
    except ValueError:
        return None, help_str
    return help_str[examples + len("Examples:") + 1 :], help_str[:examples]


class GtoCommand(GtoCliMixin, TyperCommand):
    def __init__(
        self,
        name: Optional[str],
        section: str = "other",
        aliases: List[str] = None,
        help: Optional[str] = None,
        **kwargs,
    ):
        examples, help = _extract_examples(help)
        super().__init__(
            name=name,
            section=section,
            aliases=aliases,
            examples=examples,
            help=help,
            **kwargs,
        )


class GtoGroup(GtoCliMixin, TyperGroup):
    order = [
        CommandGroups.querying,
        CommandGroups.modifying,
        CommandGroups.enriching,
        "other",
    ]

    def __init__(
        self,
        name: Optional[str] = None,
        commands: Optional[Union[Dict[str, Command], Sequence[Command]]] = None,
        section: str = "other",
        aliases: List[str] = None,
        help: str = None,
        **attrs: Any,
    ) -> None:
        examples, help = _extract_examples(help)
        super().__init__(
            name,
            help=help,
            examples=examples,
            aliases=aliases,
            section=section,
            commands=commands,
            **attrs,
        )

    def format_commands(self, ctx: Context, formatter: HelpFormatter) -> None:
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if commands:
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            sections = defaultdict(list)
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                if isinstance(cmd, (GtoCommand, GtoGroup)):
                    section = cmd.section
                    aliases = f" ({','.join(cmd.aliases)})" if cmd.aliases else ""
                else:
                    section = "other"
                    aliases = ""

                sections[section].append((subcommand + aliases, help))

            for section in self.order:
                if sections[section]:
                    with formatter.section(gettext(section)):
                        formatter.write_dl(sections[section])

    def get_command(self, ctx: Context, cmd_name: str) -> Optional[Command]:
        cmd = super().get_command(ctx, cmd_name)
        if cmd is not None:
            return cmd
        for name in self.list_commands(ctx):
            cmd = self.get_command(ctx, name)
            if (
                isinstance(cmd, (GtoCommand, GtoGroup))
                and cmd.aliases
                and cmd_name in cmd.aliases
            ):
                return cmd
        return None


def GTOGroupSection(section):
    return partial(GtoGroup, section=section)


app = Typer(
    cls=GtoGroup,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

# General Typer arguments and options
arg_name = Argument(..., help="Artifact name")
arg_version = Argument(..., help="Artifact version")
arg_stage = Argument(..., help="Stage to assign")
option_version = Option(None, "--version", help="Version to register")
option_stage = Option(None, "--stage", help="Stage to assign")
option_to_version = Option(
    None, "--to-version", help="Version to use for stage assignment"
)
option_delete = Option(
    False,
    "-d",
    "--delete",
    is_flag=True,
    help="Delete the git tag(s) instead of creating the new one",
)

# Typer options to control git-related operations
option_rev = Option(None, "--rev", help="Repo revision to use", show_default=True)
option_repo = Option(
    ".",
    "-r",
    "--repo",
    help="Local or remote repository",
    show_default=True,
)
option_all_branches = Option(
    False,
    "-a",
    "--all-branches",
    is_flag=True,
    help="Read heads from all branches",
)
option_all_commits = Option(
    False, "-A", "--all-commits", is_flag=True, help="Read all commits"
)
option_message = Option(
    None, "--message", "-m", help="Message to annotate git tag with"
)
option_force = Option(
    False,
    "--force",
    is_flag=True,
    help="Create a git tag even if it already exists and is in effect",
)


def callback_simple(  # pylint: disable=inconsistent-return-statements
    ctx: typer.Context,
    param: typer.CallbackParam,  # pylint: disable=unused-argument
    value: str,
):
    if ctx.resilient_parsing:
        return
    allowed_values = ["auto", "true", "false"]
    if value not in allowed_values:
        raise typer.BadParameter(f"Only one of {allowed_values} is allowed")
    return {"auto": None, "true": True, "false": False}[value]


option_simple = Option(
    "auto",
    "--simple",
    help="[auto, true, false] Use simple notation, e.g. rf#prod instead of rf#prod-5",
    callback=callback_simple,
)


# Typer options to control and filter the output
def callback_sort(  # pylint: disable=inconsistent-return-statements
    ctx: typer.Context,
    param: typer.CallbackParam,  # pylint: disable=unused-argument
    value: str,
):
    if ctx.resilient_parsing:
        return
    allowed_values = ["timestamp", "semver"]
    if value not in allowed_values:
        raise typer.BadParameter(f"Only one of {allowed_values} is allowed")
    return VersionSort.Timestamp if value == "timestamp" else VersionSort.SemVer


option_all = Option(False, "--all", "-a", help="Return all versions sorted")
option_registered_only = Option(
    False,
    "--registered-only",
    "--ro",
    is_flag=True,
    help="Show only registered versions",
)
option_expected = Option(
    False,
    "-e",
    "--expected",
    is_flag=True,
    help="Return exit code 1 if no result",
    show_default=True,
)
option_sort = Option(
    "timestamp",
    "--sort",
    help="Order assignments by timestamp or semver",
    callback=callback_sort,
)
option_assignments_per_version = Option(
    ASSIGNMENTS_PER_VERSION,
    "--av",
    "--assignments-per-version",
    help="Show N last stages for each version. -1 for all",
)
option_versions_per_stage = Option(
    VERSIONS_PER_STAGE,
    "--vs",
    "--versions-per-stage",
    help="Show N last versions for each stage. -1 for all. Applied after 'assignments-per-version'",
)

# Typer options to format the output
option_ascending = Option(
    False,
    "--ascending",
    "--asc",
    help="Show new first",
    show_default=True,
)
option_show_name = Option(False, "--name", is_flag=True, help="Show artifact name")
option_show_version = Option(
    False, "--version", is_flag=True, help="Output artifact version"
)
option_show_event = Option(False, "--event", is_flag=True, help="Show event")
option_show_stage = Option(False, "--stage", is_flag=True, help="Show artifact stage")
option_show_type = Option(
    False, "--type", is_flag=True, help="Show type", show_default=True
)
option_show_path = Option(
    False, "--path", is_flag=True, help="Show path", show_default=True
)
option_show_ref = Option(
    False, "--ref", is_flag=True, help="Show ref", show_default=True
)
option_show_description = Option(
    False, "--description", is_flag=True, help="Show description", show_default=True
)
option_json = Option(
    False,
    "--json",
    is_flag=True,
    help="Print output in json format",
    show_default=True,
)
option_plain = Option(
    False,
    "--plain",
    is_flag=True,
    help="Print table in grep-able format",
    show_default=True,
)
option_table = Option(
    False,
    "--table",
    is_flag=True,
    help="Print output in table format",
    show_default=True,
)
option_push_tag = Option(
    False,
    "--push",
    is_flag=True,
    help="Push created tag automatically (experimental)",
)

option_commit = Option(
    False,
    "--commit",
    is_flag=True,
    help="Automatically commit changes due to this command (experimental)",
)

option_push_commit = Option(
    False,
    "--push",
    is_flag=True,
    help="Push created commit automatically (experimental) - will set commit=True",
)


@app.callback("gto", invoke_without_command=True, no_args_is_help=True)
def gto_callback(
    ctx: Context,
    show_version: bool = Option(False, "--version", help="Show version and exit"),
    verbose: bool = Option(False, "--verbose", "-v", help="Print debug messages"),
    traceback: bool = Option(False, "--traceback", "--tb", hidden=True),
):
    """\b
    Git Tag Ops. Turn your Git repository into an Artifact Registry:
    * Registry: Track new artifacts and their versions for releases and significant
    changes.
    * Lifecycle Management: Create actionable stages for versions marking status of
    artifact or it's readiness to be consumed by a specific environment.
    * GitOps: Signal CI/CD automation or other downstream systems to act upon these
    new versions and lifecycle updates.
    * Enrichments: Annotate and query artifact metadata with additional information.
    """
    if ctx.invoked_subcommand is None and show_version:
        with cli_echo():
            echo(f"{EMOJI_GTO} GTO Version: {gto.__version__}")
    if verbose:
        logger = logging.getLogger("gto")
        logger.handlers[0].setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
        from gto.config import (  # pylint: disable=import-outside-toplevel
            CONFIG,
        )

        echo(CONFIG.__repr_str__("\n"))
        echo()

    ctx.obj = {"traceback": traceback}


def gto_command(*args, section="other", aliases=None, parent=app, **kwargs):
    def decorator(f):
        if len(args) > 0:
            cmd_name = args[0]
        else:
            cmd_name = kwargs.get("name", f.__name__)

        @parent.command(
            *args,
            **kwargs,
            cls=partial(GtoCommand, section=section, aliases=aliases),
        )
        @wraps(f)
        @click.pass_context
        def inner(ctx, *iargs, **ikwargs):
            res = {}
            error = None
            try:
                with cli_echo():
                    res = f(*iargs, **ikwargs) or {}
                res = {f"cmd_{cmd_name}_{k}": v for k, v in res.items()}
            except (ClickException, Exit, Abort) as e:
                error = str(type(e))
                raise
            except GTOException as e:
                error = str(type(e))
                if ctx.obj["traceback"]:
                    raise
                with stderr_echo():
                    echo(EMOJI_FAIL + color(str(e), col=typer.colors.RED))
                raise typer.Exit(1) from e
            except Exception as e:  # pylint: disable=broad-except
                error = str(type(e))
                if ctx.obj["traceback"]:
                    raise
                with stderr_echo():
                    echo(
                        (
                            EMOJI_FAIL
                            + color(f"Unexpected error: {str(e)}", col=typer.colors.RED)
                        )
                    )
                    echo(
                        "Please report it here running with '--traceback' flag: <https://github.com/iterative/gto/issues>"
                    )
                raise typer.Exit(1) from e
            finally:
                # TODO: analytics
                error  # pylint: disable=pointless-statement
                # send_cli_call(cmd_name, error_msg=error, **res)

        return inner

    return decorator


@gto_command(section=CommandGroups.enriching)
def annotate(
    repo: str = option_repo,
    name: str = arg_name,
    type: str = Option(None, help="Artifact type"),
    path: str = Option(None, help="Artifact path"),
    must_exist: bool = Option(
        False,
        "-e",
        "--must-exist",
        is_flag=True,
        help="Verify artifact is committed to Git",
    ),
    label: List[str] = Option(None, "--label", help="Labels to add to artifact"),
    description: str = Option("", "-d", "--description", help="Artifact description"),
    commit: bool = option_commit,
    push: bool = option_push_commit,
    # update: bool = Option(
    #     False, "-u", "--update", is_flag=True, help="Update artifact if it exists"
    # ),
):
    """Update artifact metadata annotations

    Examples:
       $ gto annotate nn --type model --path models/neural_network.h5
    """
    gto.api.annotate(
        repo,
        name,
        type=type,
        path=path,
        must_exist=must_exist,
        labels=label,
        description=description,
        commit=commit,
        push=push,
        # update=update,
    )


@gto_command(section=CommandGroups.enriching)
def remove(
    repo: str = option_repo,
    name: str = arg_name,
    commit: bool = option_commit,
    push: bool = option_push_commit,
):
    """Remove the enrichment for given artifact

    Examples:
         $ gto remove nn
    """
    gto.api.remove(repo, name, commit, push)


@gto_command(section=CommandGroups.modifying)
def register(
    repo: str = option_repo,
    name: str = arg_name,
    ref: str = Argument("HEAD", help="Git reference to use for registration"),
    version: Optional[str] = Option(
        None, "--version", "--ver", help="Version name in SemVer format"
    ),
    message: Optional[str] = option_message,
    simple: str = option_simple,
    force: bool = option_force,
    bump_major: bool = Option(
        False, "--bump-major", is_flag=True, help="Bump major version"
    ),
    bump_minor: bool = Option(
        False, "--bump-minor", is_flag=True, help="Bump minor version"
    ),
    bump_patch: bool = Option(
        False, "--bump-patch", is_flag=True, help="Bump patch version"
    ),
    push: bool = option_push_tag,
):
    """Create an artifact version to signify an important, published or released iteration

    Examples:
        Register new version at HEAD:
        $ gto register nn

        Register new version at a specific ref:
        $ gto register nn abc1234

        Assign version name explicitly:
        $ gto register nn --version v1.0.0

        Choose a part to bump version by:
        $ gto register nn --bump-minor
    """
    gto.api.register(
        repo=repo,
        name=name,
        ref=ref or "HEAD",
        version=version,
        message=message,
        simple=simple,  # type: ignore
        force=force,
        bump_major=bump_major,
        bump_minor=bump_minor,
        bump_patch=bump_patch,
        push=push,
        stdout=True,
    )


@gto_command(section=CommandGroups.modifying, aliases=["promote"])
def assign(
    repo: str = option_repo,
    name: str = arg_name,
    ref: Optional[str] = Argument(None, help="Git reference to use"),
    version: Optional[str] = Option(
        None,
        "--version",
        help="If you provide REF, this will be used to name new version",
    ),
    stage: str = option_stage,
    message: Optional[str] = option_message,
    simple: str = option_simple,
    force: bool = option_force,
    push: bool = option_push_tag,
    skip_registration: bool = Option(
        False,
        "--sr",
        "--skip-registration",
        is_flag=True,
        help="Don't register a version at specified commit",
    ),
):
    """Assign stage to specific artifact version

    Examples:
        Assign "nn" to "prod" at specific ref:
        $ gto assign nn abcd123 --stage prod

        Assign specific version:
        $ gto assign nn --version v1.0.0 --stage prod

        Assign at specific ref and name version explicitly:
        $ gto assign nn abcd123 --version v1.0.0 --stage prod

        Assign without increment:
        $ gto assign nn HEAD --stage prod --simple
    """
    if ref is not None:
        name_version = version
        version = None
    elif version is not None:
        name_version = None
    else:
        ref = "HEAD"
        name_version = None
        version = None
    gto.api.assign(
        repo,
        name,
        stage,
        version,
        ref,
        name_version,
        message=message,
        simple=simple,  # type: ignore
        force=force,
        push=push,
        skip_registration=skip_registration,
        stdout=True,
    )


@gto_command(section=CommandGroups.modifying)
def deprecate(
    repo: str = option_repo,
    name: str = arg_name,
    version: str = Argument(None, help="Artifact version"),
    stage: str = Argument(None, help="Stage to unassign"),
    message: Optional[str] = option_message,
    simple: str = option_simple,
    force: bool = option_force,
    delete: bool = option_delete,
    push: bool = option_push_tag,
):
    """Deprecate artifact, deregister a version, or unassign a stage

    Examples:
        Deprecate an artifact:
        $ gto deprecate nn

        Deprecate a version:
        $ gto deprecate nn v0.0.1

        Unassign a stage:
        $ gto deprecate nn v0.0.1 prod
    """
    if stage:
        gto.api.unassign(
            repo=repo,
            name=name,
            version=version,
            stage=stage,
            message=message,
            simple=simple,  # type: ignore
            force=force,
            delete=delete,
            push=push,
            stdout=True,
        )
    elif version:
        gto.api.deregister(
            repo=repo,
            name=name,
            version=version,
            message=message,
            simple=simple,  # type: ignore
            force=force,
            delete=delete,
            push=push,
            stdout=True,
        )
    else:
        gto.api.deprecate(
            repo=repo,
            name=name,
            message=message,
            simple=simple,  # type: ignore
            force=force,
            delete=delete,
            push=push,
            stdout=True,
        )


@gto_command(hidden=True)
def parse_tag(
    name: str = arg_name,
    key: Optional[str] = Option(None, "--key", help="Which key to return"),
):
    """Given git tag name created by this tool, parse it and return it's parts

    Examples:
        $ gto parse-tag rf@v1.0.0
        $ gto parse-tag rf#prod
    """
    parsed = gto.api.parse_tag(name)
    if key:
        parsed = parsed[key]
    format_echo(parsed, "json")


@gto_command(section=CommandGroups.querying)
def check_ref(
    repo: str = option_repo,
    ref: str = Argument(..., help="Git reference to analyze"),
    json: bool = option_json,
    name: bool = option_show_name,
    version: bool = option_show_version,
    event: bool = option_show_event,
    stage: bool = option_show_stage,
):
    """Find out the artifact version registered/assigned with ref

    Examples:
        $ gto check-ref rf@v1.0.0
        $ gto check-ref rf#prod --name
        $ gto check-ref rf#prod --version
    """
    assert (
        sum(bool(i) for i in (json, event, name, version, stage)) <= 1
    ), "Only one output formatting flags is allowed"
    result = gto.api.check_ref(repo, ref)
    if len(result) > 1:
        NotImplementedInGTO("Checking refs that created 1+ events is not supported")
    if len(result) == 0:
        return
    found_event = result[0]
    if json:
        format_echo(found_event.dict_state(exclude={"priority", "addition"}), "json")
    elif name:
        format_echo(found_event.artifact, "line")
    elif version:
        if hasattr(found_event, "version"):
            format_echo(found_event.version, "line")
    elif stage:
        if hasattr(found_event, "stage"):
            format_echo(found_event.stage, "line")
    elif event:
        format_echo(found_event.event, "line")
    else:
        echo(f"{EMOJI_OK} {found_event}")


@gto_command(section=CommandGroups.querying)
def show(  # pylint: disable=too-many-locals
    repo: str = option_repo,
    name: str = Argument(None, help="Artifact name to show. If empty, show registry"),
    all_branches: bool = option_all_branches,
    all_commits: bool = option_all_commits,
    json: bool = option_json,
    plain: bool = option_plain,
    show_name: bool = option_show_name,
    show_version: bool = option_show_version,
    show_stage: bool = option_show_stage,
    show_ref: bool = option_show_ref,
    registered_only: bool = option_registered_only,
    assignments_per_version: int = option_assignments_per_version,
    versions_per_stage: int = option_versions_per_stage,
    sort: str = option_sort,
):
    """Show the registry state, highest version, or what's assigned in stage

    Examples:
        Show the registry:
        $ gto show

        Show versions of specific artifact in registry:
        $ gto show nn

        Show greatest version or what's in stage:
        $ gto show nn@greatest
        $ gto show nn#prod
    """
    show_options = [show_name, show_version, show_stage, show_ref]
    assert (
        sum(bool(i) for i in [json, plain] + show_options) <= 1
    ), "Only one output format allowed"
    if json:
        output = gto.api.show(
            repo,
            name=name,
            all_branches=all_branches,
            all_commits=all_commits,
            registered_only=registered_only,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
            sort=sort,
            table=False,
        )
        format_echo(output, "json")
    else:
        output = gto.api.show(
            repo,
            name=name,
            all_branches=all_branches,
            all_commits=all_commits,
            registered_only=registered_only,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
            sort=sort,
            table=True,
            truncate_hexsha=True,
        )
        arg = None
        arg = "name" if show_name else arg
        arg = "version" if show_version else arg
        arg = "stage" if show_stage else arg
        arg = "ref" if show_ref else arg
        if arg:
            if arg not in output[0][0]:
                raise WrongArgs(f"Cannot apply --{arg}")
            format_echo(
                [v[arg] if isinstance(v, dict) else v for v in output[0]], "lines"
            )
        else:
            format_echo(
                output,
                format="table",
                format_table="plain" if plain else "fancy_outline",
                if_empty="Nothing found in the current workspace",
            )


@gto_command(section=CommandGroups.querying)
def history(
    repo: str = option_repo,
    name: str = Argument(None, help="Artifact name to show. If empty, show all."),
    all_branches: bool = option_all_branches,
    all_commits: bool = option_all_commits,
    json: bool = option_json,
    plain: bool = option_plain,
    ascending: bool = option_ascending,
):
    """Show a journal of registry operations

    Examples:
        $ gto history nn

        Use --all-branches and --all-commits to read more than just HEAD:
        $ gto history nn --all-commits
    """
    assert sum(bool(i) for i in (json, plain)) <= 1, "Only one output format allowed"
    if json:
        format_echo(
            gto.api.history(
                repo,
                name,
                all_branches=all_branches,
                all_commits=all_commits,
                ascending=ascending,
                table=False,
            ),
            format="json",
        )
    else:
        format_echo(
            gto.api.history(
                repo,
                name,
                all_branches=all_branches,
                all_commits=all_commits,
                ascending=ascending,
                table=True,
                truncate_hexsha=True,
            ),
            format="table",
            format_table="plain" if plain else "fancy_outline",
            if_empty="Nothing found in the current workspace",
        )


@gto_command(section=CommandGroups.querying)
def stages(
    repo: str = option_repo,
    allowed: bool = Option(
        False,
        "--allowed",
        is_flag=True,
        help="Show allowed stages from config",
        show_default=True,
    ),
    used: bool = Option(
        False,
        "--used",
        is_flag=True,
        help="Show stages that were ever used (from all git tags)",
        show_default=True,
    ),
    json: bool = option_json,
):
    """Print list of stages used in the registry

    Examples:
        $ gto stages
        $ gto stages --allowed
    """
    result = gto.api.get_stages(repo, allowed=allowed, used=used)
    if json:
        format_echo(result, "json")
    else:
        format_echo(result, "lines")


@gto_command(hidden=True)
def print_state(repo: str = option_repo):
    """Technical cmd: Print current registry state

    Examples:
        $ gto print-state
    """
    state = make_ready_to_serialize(
        gto.api._get_state(repo).dict()  # pylint: disable=protected-access
    )
    format_echo(state, "json")


@gto_command(hidden=True)
def print_index(repo: str = option_repo):
    """Technical cmd: Print repo index

    Examples:
        $ gto print-index
    """
    index = gto.api._get_index(  # pylint: disable=protected-access
        repo
    ).artifact_centric_representation()
    format_echo(index, "json")


@gto_command(section=CommandGroups.enriching)
def describe(
    repo: str = option_repo,
    name: str = arg_name,
    rev: str = option_rev,
    type: Optional[bool] = option_show_type,
    path: Optional[bool] = option_show_path,
    description: Optional[bool] = option_show_description,
):
    """Display enrichments for an artifact

    Examples:
        $ gto describe nn --rev HEAD
        $ gto describe nn@v0.0.1
    """
    assert (
        sum(bool(i) for i in (type, path, description)) <= 1
    ), "Can output one key only"
    infos = gto.api.describe(repo=repo, name=name, rev=rev)
    if not infos:
        return
    d = infos[0].get_object().dict(exclude_defaults=True)
    if type:
        if "type" not in d:
            raise WrongArgs("No type in enrichment")
        echo(d["type"])
    elif path:
        if "path" not in d:
            raise WrongArgs("No path in enrichment")
        echo(d["path"])
    elif description:
        if "description" not in d:
            raise WrongArgs("No description in enrichment")
        echo(d["description"])
    else:
        format_echo(d, "json")


if __name__ == "__main__":
    app()



import re
from collections import OrderedDict
from typing import List, Optional, Union

from funcy import distinct
from git import Repo

from gto.commit_message_generator import (
    generate_annotate_commit_message,
    generate_remove_commit_message,
)
from gto.constants import (
    ARTIFACT,
    ASSIGNMENTS_PER_VERSION,
    COMMIT,
    NAME,
    STAGE,
    VERSION,
    VERSIONS_PER_STAGE,
    VersionSort,
    mark_artifact_unregistered,
    shortcut_regexp,
)
from gto.exceptions import NoRepo, NotImplementedInGTO, WrongArgs
from gto.ext import EnrichmentInfo
from gto.git_utils import (
    clone_on_remote_repo,
    commit_produced_changes_on_commit,
    push_on_push,
    set_push_on_remote_repo,
)
from gto.index import (
    EnrichmentManager,
    FileIndexManager,
    RepoIndexManager,
    init_index_manager,
)
from gto.registry import GitRegistry
from gto.tag import NAME_REFERENCE
from gto.tag import parse_name as parse_tag_name
from gto.tag import parse_name_reference


def _is_gto_repo(repo: Union[str, Repo]):
    """Check if repo is a gto repo"""
    try:
        return GitRegistry.from_repo(repo).is_gto_repo()
    except NoRepo:
        return False


def _get_index(repo: Union[str, Repo], file=False):
    """Get index state"""
    if file:
        return FileIndexManager.from_path(
            path=repo if isinstance(repo, str) else repo.working_dir
        )
    return RepoIndexManager.from_repo(repo)


def _get_state(repo: Union[str, Repo]):
    """Show current registry state"""
    return GitRegistry.from_repo(repo).get_state()


@clone_on_remote_repo
def get_stages(repo: Union[str, Repo], allowed: bool = False, used: bool = False):
    return GitRegistry.from_repo(repo).get_stages(allowed=allowed, used=used)


# TODO: make this work the same as CLI version
@set_push_on_remote_repo
@clone_on_remote_repo
@push_on_push
@commit_produced_changes_on_commit(message_generator=generate_annotate_commit_message)
def annotate(
    repo: Union[str, Repo],
    name: str,
    type: Optional[str] = None,
    path: Optional[str] = None,
    must_exist: bool = False,
    labels: List[str] = None,
    description: str = "",
    commit: bool = False,  # pylint: disable=unused-argument
    push: bool = False,  # pylint: disable=unused-argument
    # update: bool = False,
):
    """Add an artifact to the Index"""
    return init_index_manager(path=repo).add(
        name,
        type=type,
        path=path,
        must_exist=must_exist,
        labels=labels,
        description=description,
        update=True,
    )


@set_push_on_remote_repo
@clone_on_remote_repo
@push_on_push
@commit_produced_changes_on_commit(message_generator=generate_remove_commit_message)
def remove(
    repo: Union[str, Repo],
    name: str,
    commit: bool = False,
    push: bool = False,
):  # pylint: disable=unused-argument
    """Remove an artifact from the Index"""
    return init_index_manager(path=repo).remove(name)


@set_push_on_remote_repo
@clone_on_remote_repo
def register(
    repo: Union[str, Repo],
    name: str,
    ref: str,
    version: str = None,
    message: str = None,
    simple: bool = None,
    force: bool = False,
    bump_major: bool = False,
    bump_minor: bool = False,
    bump_patch: bool = False,
    push: bool = False,
    stdout: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    """Register new artifact version"""
    return GitRegistry.from_repo(repo).register(
        name=name,
        ref=ref,
        version=version,
        message=message,
        simple=simple if simple is not None else True,
        force=force,
        bump_major=bump_major,
        bump_minor=bump_minor,
        bump_patch=bump_patch,
        push=push,
        stdout=stdout,
        author=author,
        author_email=author_email,
    )


@set_push_on_remote_repo
@clone_on_remote_repo
def assign(
    repo: Union[str, Repo],
    name: str,
    stage: str,
    version: Optional[str] = None,
    ref: Optional[str] = None,
    name_version: Optional[str] = None,
    message: Optional[str] = None,
    simple: bool = False,
    force: bool = False,
    push: bool = False,
    skip_registration: bool = False,
    stdout: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    """Assign stage to specific artifact version"""
    return GitRegistry.from_repo(repo).assign(
        name=name,
        stage=stage,
        version=version,
        ref=ref,
        name_version=name_version,
        message=message,
        simple=simple,
        force=force,
        push=push,
        skip_registration=skip_registration,
        stdout=stdout,
        author=author,
        author_email=author_email,
    )


@set_push_on_remote_repo
@clone_on_remote_repo
def unassign(
    repo: Union[str, Repo],
    name: str,
    ref: str = None,
    version: str = None,
    stage: str = None,
    message: str = None,
    stdout: bool = False,
    simple: Optional[bool] = None,
    force: bool = False,
    delete: bool = False,
    push: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    return GitRegistry.from_repo(repo).unassign(
        name=name,
        stage=stage,
        ref=ref,
        version=version,
        message=message,
        stdout=stdout,
        simple=simple if simple is not None else False,
        force=force,
        delete=delete,
        push=push,
        author=author,
        author_email=author_email,
    )


@set_push_on_remote_repo
@clone_on_remote_repo
def deregister(
    repo: Union[str, Repo],
    name: str,
    ref: str = None,
    version: str = None,
    message: str = None,
    stdout: bool = False,
    simple: Optional[bool] = None,
    force: bool = False,
    delete: bool = False,
    push: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    return GitRegistry.from_repo(repo).deregister(
        name=name,
        ref=ref,
        version=version,
        message=message,
        stdout=stdout,
        simple=simple if simple is not None else True,
        force=force,
        delete=delete,
        push=push,
        author=author,
        author_email=author_email,
    )


@set_push_on_remote_repo
@clone_on_remote_repo
def deprecate(
    repo: Union[str, Repo],
    name: str,
    message: str = None,
    stdout: bool = False,
    simple: Optional[bool] = None,
    force: bool = False,
    delete: bool = False,
    push: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    return GitRegistry.from_repo(repo).deprecate(
        name=name,
        message=message,
        stdout=stdout,
        simple=simple if simple is not None else True,
        force=force,
        delete=delete,
        push=push,
        author=author,
        author_email=author_email,
    )


def parse_tag(name: str):
    return parse_tag_name(name)


def find_latest_version(
    repo: Union[str, Repo],
    name: str,
    all: bool = False,
    registered: bool = True,
):
    """Return latest version for artifact"""
    return GitRegistry.from_repo(repo).latest(name, all=all, registered=registered)


def find_versions_in_stage(
    repo: Union[str, Repo],
    name: str,
    stage: str,
    assignments_per_version=ASSIGNMENTS_PER_VERSION,
    versions_per_stage=VERSIONS_PER_STAGE,
    registered_only: bool = False,
):
    """Return version of artifact with specific stage active"""
    return GitRegistry.from_repo(repo).which(
        name,
        stage,
        raise_if_not_found=False,
        assignments_per_version=assignments_per_version,
        versions_per_stage=versions_per_stage,
        registered_only=registered_only,
    )


@clone_on_remote_repo
def check_ref(repo: Union[str, Repo], ref: str):
    """Find out what have been registered/assigned in the provided ref"""
    reg = GitRegistry.from_repo(repo)
    return reg.check_ref(ref)


@clone_on_remote_repo
def show(
    repo: Union[str, Repo],
    name: Optional[str] = None,
    all_branches=False,
    all_commits=False,
    truncate_hexsha=False,
    registered_only=False,
    assignments_per_version=ASSIGNMENTS_PER_VERSION,
    versions_per_stage=VERSIONS_PER_STAGE,
    sort=VersionSort.Timestamp,
    table: bool = False,
):
    return (
        _show_versions(
            repo,
            name=name,
            all_branches=all_branches,
            all_commits=all_commits,
            registered_only=registered_only,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
            sort=sort,
            table=table,
            truncate_hexsha=truncate_hexsha,
        )
        if name
        else _show_registry(
            repo,
            all_branches=all_branches,
            all_commits=all_commits,
            registered_only=registered_only,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
            sort=sort,
            table=table,
            truncate_hexsha=truncate_hexsha,
        )
    )


def _show_registry(
    repo: Union[str, Repo],
    all_branches=False,
    all_commits=False,
    registered_only=False,
    assignments_per_version: int = None,
    versions_per_stage: int = None,
    sort: VersionSort = None,
    table: bool = False,
    truncate_hexsha: bool = False,
):
    """Show current registry state"""

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha else hexsha

    reg = GitRegistry.from_repo(repo)
    stages = list(reg.get_stages())
    models_state = {
        o.artifact: {
            "version": format_hexsha(o.get_latest_version(registered_only=True).version)
            if o.get_latest_version(registered_only=True)
            else None,
            "stage": {
                name: ", ".join(
                    [
                        format_hexsha(s.version)
                        for s in o.get_vstages(
                            registered_only=registered_only,
                            assignments_per_version=assignments_per_version,
                            versions_per_stage=versions_per_stage,
                            sort=sort,
                        ).get(name, [])
                    ]
                )
                or None
                for name in stages
            },
            "registered": o.is_registered,
        }
        for o in reg.get_artifacts(
            all_branches=all_branches,
            all_commits=all_commits,
        ).values()
    }
    if not table:
        return models_state

    return [
        OrderedDict(
            zip(
                ["name", "latest"] + [f"#{e}" for e in stages],
                [
                    name if d["registered"] else mark_artifact_unregistered(name),
                    d["version"],
                ]
                + [d["stage"][name] for name in stages],
            ),
        )
        for name, d in models_state.items()
    ], "keys"


def _show_versions(  # pylint: disable=too-many-locals
    repo: Union[str, Repo],
    name: str,
    raw: bool = False,
    all_branches=False,
    all_commits=False,
    registered_only=False,
    assignments_per_version: int = None,
    versions_per_stage: int = None,
    sort: VersionSort = None,
    table: bool = False,
    truncate_hexsha: bool = False,
):
    """List versions of artifact"""

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha else hexsha

    match = re.search(shortcut_regexp, name)
    if match:
        name = match["artifact"]

    reg = GitRegistry.from_repo(repo)
    if raw:
        return reg.find_artifact(name).versions

    artifact = reg.find_artifact(
        name,
        all_branches=all_branches,
        all_commits=all_commits,
    )
    stages = artifact.get_vstages(
        registered_only=registered_only,
        assignments_per_version=assignments_per_version,
        versions_per_stage=versions_per_stage,
        sort=sort,
    )
    versions = []
    for v in artifact.get_versions(
        include_non_explicit=not registered_only, include_discovered=True
    ):
        v = v.dict_state()
        v["stages"] = [
            vstage.dict_state()
            for vstages in stages.values()
            for vstage in vstages
            if vstage.version == v["version"]
        ]
        versions.append(v)

    if match and match["greatest"]:
        versions = versions[:1]
    if match and match["stage"]:
        versions = [
            v for v in versions for a in v["stages"] if match["stage"] in a["stage"]
        ]

    if not table:
        return versions

    first_keys = ["artifact", "version", "stage"]
    versions_ = []
    for v in versions:
        v["artifact"] = (
            v["artifact"]
            if artifact.is_registered
            else mark_artifact_unregistered(v["artifact"])
        )
        v["version"] = format_hexsha(v["version"])
        v["stage"] = ", ".join(
            distinct(  # TODO: remove? no longer necessary
                s["stage"] for s in v["stages"]
            )
        )
        v["commit_hexsha"] = format_hexsha(v["commit_hexsha"])
        if len(v["registrations"]) > 1:
            raise NotImplementedInGTO(
                "Multiple registrations are not supported currently. How you got in here?"
            )
        for key in list(v.keys()):
            if key not in first_keys + ["created_at", "ref"]:
                del v[key]
        v = OrderedDict(
            [(key, v[key]) for key in first_keys]
            + [(key, v[key]) for key in v if key not in first_keys]
        )
        versions_.append(v)
    return versions_, "keys"


@clone_on_remote_repo
def describe(
    repo: Union[str, Repo], name: str, rev: str = None
) -> List[EnrichmentInfo]:
    """Find enrichments for the artifact"""
    ref_type, parsed = parse_name_reference(name)
    if ref_type == NAME_REFERENCE.NAME:
        return EnrichmentManager.from_repo(repo).describe(name=name, rev=rev)
    if ref_type == NAME_REFERENCE.TAG:
        if rev:
            raise WrongArgs("Should not specify revision if you pass git tag")
        return EnrichmentManager.from_repo(repo).describe(name=parsed[NAME], rev=name)
    raise NotImplementedError


@clone_on_remote_repo
def history(
    repo: Union[str, Repo],
    artifact: str = None,
    # action: str = None,
    all_branches=False,
    all_commits=False,
    ascending: bool = False,
    table: bool = False,
    truncate_hexsha: bool = False,
):

    reg = GitRegistry.from_repo(repo)
    artifacts = reg.get_artifacts(
        all_branches=all_branches,
        all_commits=all_commits,
    )

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha else hexsha

    events = [
        OrderedDict(
            timestamp=e.created_at,
            artifact=e.artifact,
            event=type(e).__name__.lower(),
            priority=e.priority,
            version=format_hexsha(e.version) if hasattr(e, "version") else None,
            stage=getattr(e, "stage", None),
            commit=format_hexsha(e.commit_hexsha),
            author=e.author,
            author_email=e.author_email,
            message=e.message,
            ref=format_hexsha(e.ref) if e.ref == e.commit_hexsha else e.ref,
        )
        for o in artifacts.values()
        for e in o.get_events()
    ]

    events = sorted(
        events,
        key=lambda x: (x["timestamp"], x["priority"]),
    )
    if not ascending:
        events.reverse()
    if artifact:
        events = [event for event in events if event["artifact"] == artifact]
    if not table:
        return events
    keys_order = [
        "timestamp",
        ARTIFACT,
        "event",
        VERSION,
        STAGE,
        COMMIT,
        "ref",
    ]
    keys_order = [c for c in keys_order if any(c in event for event in events)]
    events = [
        OrderedDict((key, event.get(key)) for key in keys_order) for event in events
    ]
    return events, "keys"



# pylint: disable=unused-variable, protected-access
"""TODO: add more tests for API"""
import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, Optional, Tuple
from unittest.mock import call, patch

import git
import pytest
from freezegun import freeze_time

import gto
import tests.resources
from gto.api import show
from gto.commit_message_generator import (
    generate_annotate_commit_message,
    generate_remove_commit_message,
)
from gto.exceptions import PathIsUsed, WrongArgs
from gto.git_utils import git_clone
from gto.tag import find
from gto.versions import SemVer
from tests.skip_presets import skip_for_windows_py_lt_3_9
from tests.utils import (
    check_obj,
    convert_objects_to_str_in_json_serializable_object,
)


def test_empty_index(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo
    index = gto.api._get_index(repo.working_dir)
    assert len(index.artifact_centric_representation()) == 0


def test_empty_state(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo
    state = gto.api._get_state(repo.working_dir)
    assert len(state.artifacts) == 0


def test_api_info_commands_empty_repo(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo
    gto.api.show(repo.working_dir)
    gto.api.history(repo.working_dir)


def test_add_remove(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo
    name, type, path, must_exist = "new-artifact", "new-type", "new/path", False
    gto.api.annotate(
        repo.working_dir, name, type=type, path=path, must_exist=must_exist
    )
    with pytest.raises(PathIsUsed):
        gto.api.annotate(repo.working_dir, "other-name", path=path)
    index = gto.api._get_index(repo.working_dir).get_index()
    assert name in index
    check_obj(
        index.state[name],
        dict(
            type=type,
            path=path,
            virtual=not must_exist,
            labels=[],
            description="",
        ),
        [],
    )
    gto.api.remove(repo.working_dir, name)
    index = gto.api._get_index(repo.working_dir).get_index()
    assert name not in index


@pytest.fixture
def repo_with_artifact(init_showcase_semver):
    repo: git.Repo
    repo, write_file = init_showcase_semver
    name, type, path, must_exist = "new-artifact", "new-type", "new/path", False
    gto.api.annotate(
        repo.working_dir, name, type=type, path=path, must_exist=must_exist
    )
    repo.index.add(["artifacts.yaml"])
    repo.index.commit("Added index")
    gto.api.annotate(
        repo.working_dir, name, type=type, path="path", must_exist=must_exist
    )
    repo.index.add(["artifacts.yaml"])
    repo.index.commit("Added index")
    return repo, name


def test_api_info_commands_repo_with_artifact(
    repo_with_artifact: Tuple[git.Repo, Callable]
):
    repo, write_file = repo_with_artifact
    gto.api.show(repo.working_dir)
    gto.api.show(repo.working_dir, "new-artifact")
    gto.api.history(repo.working_dir)


def test_register_deregister(repo_with_artifact):
    repo, name = repo_with_artifact
    vname1, vname2 = "v1.0.0", "v1.0.1"
    gto.api.register(repo.working_dir, name, "HEAD", vname1)
    latest = gto.api.find_latest_version(repo.working_dir, name)
    assert latest.version == vname1
    gto.api.annotate(
        repo.working_dir,
        "something-irrelevant",
        "doesnt-matter",
        "anything",
        must_exist=False,
    )
    repo.index.commit(
        "Irrelevant action to create a git commit to register another version"
    )
    message = "Some message"
    author = "GTO"
    author_email = "gto@iterative.ai"
    gto.api.register(
        repo.working_dir,
        name,
        "HEAD",
        message=message,
        author=author,
        author_email=author_email,
    )
    latest = gto.api.find_latest_version(repo.working_dir, name)
    assert latest.version == vname2
    assert latest.message == message
    assert latest.author == author
    assert latest.author_email == author_email

    gto.api.deregister(repo=repo.working_dir, name=name, version=vname2)
    latest = gto.api.find_latest_version(repo.working_dir, name)
    assert latest.version == vname1


def test_assign(repo_with_artifact: Tuple[git.Repo, str]):
    repo, name = repo_with_artifact
    stage = "staging"
    repo.create_tag("v1.0.0")
    repo.create_tag("wrong-tag-unrelated")
    message = "some msg"
    author = "GTO"
    author_email = "gto@iterative.ai"
    event = gto.api.assign(
        repo.working_dir,
        name,
        stage,
        ref="HEAD",
        name_version="v0.0.1",
        message=message,
        author=author,
        author_email=author_email,
    )
    assignments = gto.api.find_versions_in_stage(repo.working_dir, name, stage)
    assert len(assignments) == 1
    check_obj(
        assignments[0].dict_state(),
        dict(
            artifact=name,
            version="v0.0.1",
            stage=stage,
            author=author,
            author_email=author_email,
            message=message,
            commit_hexsha=repo.commit().hexsha,
            is_active=True,
            ref=event.ref,
        ),
        {"created_at", "assignments", "unassignments", "tag", "activated_at"},
    )


def test_assign_skip_registration(repo_with_artifact: Tuple[git.Repo, str]):
    repo, name = repo_with_artifact
    stage = "staging"
    with pytest.raises(WrongArgs):
        gto.api.assign(
            repo.working_dir,
            name,
            stage,
            ref="HEAD",
            name_version="v0.0.1",
            skip_registration=True,
        )
    gto.api.assign(repo.working_dir, name, stage, ref="HEAD", skip_registration=True)
    assignments = gto.api.find_versions_in_stage(repo.working_dir, name, stage)
    assert len(assignments) == 1
    assert not SemVer.is_valid(assignments[0].version)


def test_assign_force_is_needed(repo_with_artifact: Tuple[git.Repo, str]):
    repo, name = repo_with_artifact
    gto.api.assign(repo, name, "staging", ref="HEAD")
    gto.api.assign(repo, name, "staging", ref="HEAD^1")
    with pytest.raises(WrongArgs):
        gto.api.assign(repo, name, "staging", ref="HEAD")
    with pytest.raises(WrongArgs):
        gto.api.assign(repo, name, "staging", ref="HEAD^1")
    gto.api.assign(repo, name, "staging", ref="HEAD", force=True)
    gto.api.assign(repo, name, "staging", ref="HEAD^1", force=True)


@contextmanager
def environ(**overrides):
    old = {name: os.environ[name] for name in overrides if name in os.environ}
    to_del = set(overrides) - set(old)
    try:
        os.environ.update(overrides)
        yield
    finally:
        os.environ.update(old)
        for name in to_del:
            os.environ.pop(name, None)


def test_check_ref_detailed(repo_with_artifact: Tuple[git.Repo, Callable]):
    repo, name = repo_with_artifact

    NAME = "model"
    SEMVER = "v1.2.3"
    GIT_AUTHOR_NAME = "Alexander Guschin"
    GIT_AUTHOR_EMAIL = "aguschin@iterative.ai"
    GIT_COMMITTER_NAME = "Oliwav"
    GIT_COMMITTER_EMAIL = "oliwav@iterative.ai"

    with environ(
        GIT_AUTHOR_NAME=GIT_AUTHOR_NAME,
        GIT_AUTHOR_EMAIL=GIT_AUTHOR_EMAIL,
        GIT_COMMITTER_NAME=GIT_COMMITTER_NAME,
        GIT_COMMITTER_EMAIL=GIT_COMMITTER_EMAIL,
    ):
        gto.api.register(repo, name=NAME, ref="HEAD", version=SEMVER)

    events = gto.api.check_ref(repo, f"{NAME}@{SEMVER}")
    assert len(events) == 1, "Should return one event"
    check_obj(
        events[0].dict_state(),
        {
            "event": "registration",
            "artifact": NAME,
            "version": SEMVER,
            "author": GIT_COMMITTER_NAME,
            "author_email": GIT_COMMITTER_EMAIL,
            "tag": f"{NAME}@{SEMVER}",
        },
        skip_keys={"commit_hexsha", "created_at", "message", "priority", "addition"},
    )


def test_check_ref_multiple_showcase(showcase):
    repo: git.Repo
    (
        path,
        repo,
        write_file,
        first_commit,
        second_commit,
    ) = showcase

    for tag in find(repo=repo):
        events = gto.api.check_ref(repo, tag.name)
        assert len(events) == 1, "Should return one event"
        assert events[0].ref == tag.name


def test_check_ref_catch_the_bug(repo_with_artifact: Tuple[git.Repo, Callable]):
    repo, name = repo_with_artifact
    NAME = "artifact"
    gto.api.register(repo, NAME, "HEAD")
    assignment1 = gto.api.assign(repo, NAME, "staging", ref="HEAD")
    assignment2 = gto.api.assign(repo, NAME, "prod", ref="HEAD")
    assignment3 = gto.api.assign(repo, NAME, "dev", ref="HEAD")
    for assignment, tag in zip(
        [assignment1, assignment2, assignment3],
        [f"{NAME}#staging#1", f"{NAME}#prod#2", f"{NAME}#dev#3"],
    ):
        events = gto.api.check_ref(repo, tag)
        assert len(events) == 1, events
        assert events[0].ref == assignment.tag == tag


def test_is_not_gto_repo(empty_git_repo):
    repo, _ = empty_git_repo
    assert not gto.api._is_gto_repo(repo.working_dir)


def test_is_gto_repo_because_of_config(init_showcase_semver):
    repo, _ = init_showcase_semver
    assert gto.api._is_gto_repo(repo.working_dir)


def test_is_gto_repo_because_of_registered_artifact(repo_with_commit):
    repo, _ = repo_with_commit
    gto.api.register(repo, "model", "HEAD", "v1.0.0")
    assert gto.api._is_gto_repo(repo)


def test_is_gto_repo_because_of_artifacts_yaml(empty_git_repo):
    repo, write_file = empty_git_repo
    write_file("artifacts.yaml", "{}")
    assert gto.api._is_gto_repo(repo)


@skip_for_windows_py_lt_3_9
def test_if_show_on_remote_git_repo_then_return_expected_registry():
    result = show(repo=tests.resources.SAMPLE_REMOTE_REPO_URL)
    assert result == tests.resources.get_sample_remote_repo_expected_registry()


@skip_for_windows_py_lt_3_9
@pytest.mark.parametrize(
    "ref,expected_stage,expected_version,expected_artifact",
    (
        ("churn#prod#2", "prod", "v3.0.0", "churn"),
        ("segment@v0.4.1", None, "v0.4.1", "segment"),
    ),
)
def test_if_check_ref_on_remote_git_repo_then_return_expected_reference(
    ref: str,
    expected_stage: Optional[str],
    expected_version: str,
    expected_artifact: str,
):
    result = gto.api.check_ref(repo=tests.resources.SAMPLE_REMOTE_REPO_URL, ref=ref)
    assert len(result) == 1
    if expected_stage is not None:
        assert result[0].stage == expected_stage
    else:
        assert hasattr(result[0], "stage") is False
    assert result[0].version == expected_version
    assert result[0].artifact == expected_artifact


@freeze_time("1996-06-09 00:00:00", tz_offset=0)
@skip_for_windows_py_lt_3_9
def test_if_history_on_remote_git_repo_then_return_expected_history():
    result = gto.api.history(
        repo=tests.resources.SAMPLE_REMOTE_REPO_URL, artifact="churn"
    )
    assert (
        convert_objects_to_str_in_json_serializable_object(result)
        == tests.resources.get_sample_remote_repo_expected_history_churn()
    )


@skip_for_windows_py_lt_3_9
def test_if_stages_on_remote_git_repo_then_return_expected_stages():
    result = gto.api.get_stages(repo=tests.resources.SAMPLE_REMOTE_REPO_URL)
    assert result == ["dev", "prod", "staging"]


@skip_for_windows_py_lt_3_9
def test_if_describe_on_remote_git_repo_then_return_expected_info():
    result = gto.api.describe(repo=tests.resources.SAMPLE_REMOTE_REPO_URL, name="churn")
    assert result[0].get_object().dict(exclude_defaults=True) == {
        "type": "model",
        "path": "models/churn.pkl",
        "virtual": False,
    }


def test_if_register_with_auto_push_then_invoke_git_push_tag(repo_with_artifact):
    repo, _ = repo_with_artifact
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.register(repo=repo.working_dir, name="model", ref="HEAD", push=True)
    mocked_git_push_tags.assert_called_once_with(
        repo_path=Path(repo.working_dir).as_posix(),
        tag_name="model@v0.0.1",
        delete=False,
    )


def test_if_assign_with_auto_push_then_invoke_git_push_tag_2_times_for_registration_and_promotion(
    repo_with_artifact,
):
    repo, _ = repo_with_artifact
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.assign(
            repo.working_dir, name="model", stage="dev", ref="HEAD", push=True
        )
    expected_calls = [
        call(
            repo_path=Path(repo.working_dir).as_posix(),
            tag_name="model@v0.0.1",
            delete=False,
        ),
        call(
            repo_path=Path(repo.working_dir).as_posix(),
            tag_name="model#dev#1",
            delete=False,
        ),
    ]
    mocked_git_push_tags.assert_has_calls(expected_calls)


def test_if_unassign_with_auto_push_then_invoke_git_push_tag(repo_with_artifact):
    repo, _ = repo_with_artifact
    gto.api.assign(repo.working_dir, name="model", stage="dev", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.unassign(
            repo.working_dir,
            name="model",
            stage="dev",
            version="v0.0.1",
            push=True,
        )
    mocked_git_push_tags.assert_called_once_with(
        repo_path=Path(repo.working_dir).as_posix(),
        tag_name="model#dev!#2",
        delete=False,
    )


def test_if_unassign_with_delete_and_auto_push_then_invoke_git_push_tag(
    repo_with_artifact,
):
    repo, _ = repo_with_artifact
    gto.api.assign(repo.working_dir, name="model", stage="dev", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.unassign(
            repo.working_dir,
            name="model",
            stage="dev",
            version="v0.0.1",
            delete=True,
            push=True,
        )
    mocked_git_push_tags.assert_called_once_with(
        repo_path=Path(repo.working_dir).as_posix(), tag_name="model#dev#1", delete=True
    )


def test_if_deregister_with_auto_push_then_invoke_git_push_tag(repo_with_artifact):
    repo, _ = repo_with_artifact
    gto.api.register(repo.working_dir, name="model", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.deregister(repo.working_dir, name="model", version="v0.0.1", push=True)
    mocked_git_push_tags.assert_called_once_with(
        repo_path=Path(repo.working_dir).as_posix(),
        tag_name="model@v0.0.1!",
        delete=False,
    )


def test_if_deregister_with_delete_and_auto_push_then_invoke_git_push_tag(
    repo_with_artifact,
):
    repo, _ = repo_with_artifact
    gto.api.register(repo.working_dir, name="model", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.deregister(
            repo.working_dir,
            name="model",
            version="v0.0.1",
            push=True,
            delete=True,
        )
    mocked_git_push_tags.assert_called_once_with(
        repo_path=Path(repo.working_dir).as_posix(),
        tag_name="model@v0.0.1",
        delete=True,
    )


def test_if_deprecate_with_auto_push_then_invoke_git_push_tag(repo_with_artifact):
    repo, _ = repo_with_artifact
    gto.api.register(repo.working_dir, name="model", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.deprecate(repo.working_dir, name="model", push=True)
    mocked_git_push_tags.assert_called_once_with(
        repo_path=Path(repo.working_dir).as_posix(),
        tag_name="model@deprecated",
        delete=False,
    )


def test_if_deprecate_with_delete_and_auto_push_then_invoke_git_push_tag(
    repo_with_artifact,
):
    repo, _ = repo_with_artifact
    gto.api.register(repo.working_dir, name="model", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.deprecate(repo.working_dir, name="model", push=True, delete=True)
    mocked_git_push_tags.assert_called_once_with(
        repo_path=Path(repo.working_dir).as_posix(),
        tag_name="model@v0.0.1",
        delete=True,
    )


@skip_for_windows_py_lt_3_9
def test_if_register_with_remote_repo_then_invoke_git_push_tag():
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            # pylint: disable=consider-using-with
            tmp_dir = TemporaryDirectory()
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.register(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="model",
                ref="HEAD",
            )
            mocked_git_push_tag.assert_called_once_with(
                repo_path=Path(tmp_dir.name).as_posix(),
                tag_name="model@v0.0.1",
                delete=False,
            )
            tmp_dir.cleanup()


@skip_for_windows_py_lt_3_9
def test_if_assign_with_remote_repo_then_invoke_git_push_tag():
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            # pylint: disable=consider-using-with
            tmp_dir = TemporaryDirectory()
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.assign(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="model",
                stage="dev",
                ref="HEAD",
            )
            expected_calls = [
                call(
                    repo_path=Path(tmp_dir.name).as_posix(),
                    tag_name="model@v0.0.1",
                    delete=False,
                ),
                call(
                    repo_path=Path(tmp_dir.name).as_posix(),
                    tag_name="model#dev#1",
                    delete=False,
                ),
            ]
            mocked_git_push_tag.assert_has_calls(expected_calls)
            tmp_dir.cleanup()


@skip_for_windows_py_lt_3_9
def test_if_deprecate_with_remote_repo_then_invoke_git_push_tag():
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            # pylint: disable=consider-using-with
            tmp_dir = TemporaryDirectory()
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.deprecate(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="churn",
            )
            mocked_git_push_tag.assert_called_once_with(
                repo_path=Path(tmp_dir.name).as_posix(),
                tag_name="churn@deprecated",
                delete=False,
            )
            tmp_dir.cleanup()


@skip_for_windows_py_lt_3_9
def test_if_deregister_with_remote_repo_then_invoke_git_push_tag():
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            # pylint: disable=consider-using-with
            tmp_dir = TemporaryDirectory()
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.deregister(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="churn",
                version="v3.0.0",
            )
            mocked_git_push_tag.assert_called_once_with(
                repo_path=Path(tmp_dir.name).as_posix(),
                tag_name="churn@v3.0.0!",
                delete=False,
            )
            tmp_dir.cleanup()


@skip_for_windows_py_lt_3_9
def test_if_unassign_with_remote_repo_then_invoke_git_push_tag():
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            # pylint: disable=consider-using-with
            tmp_dir = TemporaryDirectory()
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.unassign(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="churn",
                stage="staging",
                version="v3.1.0",
            )
            mocked_git_push_tag.assert_called_once_with(
                repo_path=Path(tmp_dir.name).as_posix(),
                tag_name="churn#staging!#3",
                delete=False,
            )
            tmp_dir.cleanup()


def test_if_annotate_with_auto_commit_then_invoke_stash_and_commit(
    init_showcase_semver,
):
    repo, write_file = init_showcase_semver
    name, type, path, must_exist = "new-artifact", "new-type", "new/path", False
    repo.index.commit(message="first commit")

    with patch("gto.git_utils.stashed_changes") as mocked_stashed_changes:
        mocked_stashed_changes.return_value.__enter__.return_value = [], []
        with patch(
            "gto.git_utils.git_add_and_commit_all_changes"
        ) as mocked_git_add_and_commit_all_changes:
            gto.api.annotate(
                repo.working_dir,
                name,
                type=type,
                path=path,
                must_exist=must_exist,
                commit=True,
            )

    mocked_stashed_changes.assert_called_once_with(
        repo_path=repo.working_dir, include_untracked=True
    )
    mocked_git_add_and_commit_all_changes.assert_called_once_with(
        repo_path=repo.working_dir,
        message=generate_annotate_commit_message(name=name, type=type, path=path),
    )


def test_if_remove_with_auto_commit_then_invoke_stash_and_commit(
    init_showcase_semver,
):
    repo, write_file = init_showcase_semver
    name, type, path, must_exist = "new-artifact", "new-type", "new/path", False
    repo.index.commit(message="first commit")
    gto.api.annotate(
        repo.working_dir,
        name,
        type=type,
        path=path,
        must_exist=must_exist,
        commit=True,
    )

    with patch("gto.git_utils.stashed_changes") as mocked_stashed_changes:
        mocked_stashed_changes.return_value.__enter__.return_value = [], []
        with patch(
            "gto.git_utils.git_add_and_commit_all_changes"
        ) as mocked_git_add_and_commit_all_changes:
            gto.api.remove(repo=repo.working_dir, name=name, commit=True)

    mocked_stashed_changes.assert_called_once_with(
        repo_path=repo.working_dir, include_untracked=True
    )
    mocked_git_add_and_commit_all_changes.assert_called_once_with(
        repo_path=repo.working_dir, message=generate_remove_commit_message(name=name)
    )


def test_if_annotate_with_auto_push_then_invoke_commit_and_push(init_showcase_semver):
    repo, write_file = init_showcase_semver
    name, type, path, must_exist = "new-artifact", "new-type", "new/path", False
    repo.index.commit(message="first commit")

    with patch("gto.git_utils.stashed_changes") as mocked_stashed_changes:
        mocked_stashed_changes.return_value.__enter__.return_value = [], []
        with patch(
            "gto.git_utils.git_add_and_commit_all_changes"
        ) as mocked_git_add_and_commit_all_changes:
            with patch("gto.git_utils.git_push") as mocked_git_push:
                gto.api.annotate(
                    repo.working_dir,
                    name,
                    type=type,
                    path=path,
                    must_exist=must_exist,
                    push=True,
                )

    mocked_stashed_changes.assert_called_once_with(
        repo_path=repo.working_dir, include_untracked=True
    )
    mocked_git_add_and_commit_all_changes.assert_called_once_with(
        repo_path=repo.working_dir,
        message=generate_annotate_commit_message(name=name, type=type, path=path),
    )
    mocked_git_push.assert_called_once_with(repo_path=repo.working_dir)


def test_if_remove_with_auto_push_then_invoke_commit_and_push(
    init_showcase_semver,
):
    repo, write_file = init_showcase_semver
    name, type, path, must_exist = "new-artifact", "new-type", "new/path", False
    repo.index.commit(message="first commit")
    gto.api.annotate(
        repo.working_dir,
        name,
        type=type,
        path=path,
        must_exist=must_exist,
        commit=True,
    )

    with patch("gto.git_utils.stashed_changes") as mocked_stashed_changes:
        mocked_stashed_changes.return_value.__enter__.return_value = [], []
        with patch(
            "gto.git_utils.git_add_and_commit_all_changes"
        ) as mocked_git_add_and_commit_all_changes:
            with patch("gto.git_utils.git_push") as mocked_git_push:
                gto.api.remove(repo=repo.working_dir, name=name, push=True)

    mocked_stashed_changes.assert_called_once_with(
        repo_path=repo.working_dir, include_untracked=True
    )
    mocked_git_add_and_commit_all_changes.assert_called_once_with(
        repo_path=repo.working_dir, message=generate_remove_commit_message(name=name)
    )
    mocked_git_push.assert_called_once_with(repo_path=repo.working_dir)


@skip_for_windows_py_lt_3_9
def test_if_annotate_with_remote_repo_then_clone_and_push():
    with patch("gto.git_utils.git_push") as mocked_git_push:
        with patch("gto.git_utils.git_clone") as mocked_git_clone:
            mocked_git_clone.side_effect = git_clone
            with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
                MockedTemporaryDirectory.return_value = (
                    TemporaryDirectory()  # pylint: disable=consider-using-with
                )
                gto.api.annotate(
                    repo=tests.resources.SAMPLE_REMOTE_REPO_URL, name="test-model"
                )

    mocked_git_push.assert_called_once_with(
        repo_path=MockedTemporaryDirectory.return_value.name
    )
    mocked_git_clone.assert_called_once_with(
        repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
        dir=MockedTemporaryDirectory.return_value.name,
    )


@skip_for_windows_py_lt_3_9
def test_if_remove_with_remote_repo_then_clone_and_push():
    with patch("gto.git_utils.git_push") as mocked_git_push:
        with patch("gto.git_utils.git_clone") as mocked_git_clone:
            mocked_git_clone.side_effect = git_clone
            with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
                MockedTemporaryDirectory.return_value = (
                    TemporaryDirectory()  # pylint: disable=consider-using-with
                )
                gto.api.remove(
                    repo=tests.resources.SAMPLE_REMOTE_REPO_URL, name="segment"
                )

    mocked_git_push.assert_called_once_with(
        repo_path=MockedTemporaryDirectory.return_value.name
    )
    mocked_git_clone.assert_called_once_with(
        repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
        dir=MockedTemporaryDirectory.return_value.name,
    )



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, Tuple, Union
from unittest.mock import MagicMock, call, patch

import pytest
from git import Repo

import tests.resources
from gto.exceptions import GTOException
from gto.git_utils import (
    clone_on_remote_repo,
    commit_produced_changes_on_commit,
    git_add_and_commit_all_changes,
    git_clone,
    git_push,
    git_push_tag,
    is_url_of_remote_repo,
    push_on_push,
    set_push_on_remote_repo,
    stashed_changes,
)
from tests.skip_presets import (
    only_for_windows_py_lt_3_8,
    skip_for_windows_py_lt_3_9,
)


def test_clone_on_remote_repo_if_repo_is_a_meaningless_string_then_leave_it_unchanged():
    assert_f_called_with_repo_return_repo_itself(repo="meaningless_string")


def test_clone_on_remote_repo_if_repo_is_a_local_git_repo_then_leave_it_unchanged(
    tmp_local_empty_git_repo,
):
    assert_f_called_with_repo_return_repo_itself(repo=tmp_local_empty_git_repo)


def test_clone_on_remote_repo_if_repo_gitpython_object_then_leave_it_unchanged(
    tmp_local_empty_git_repo,
):
    assert_f_called_with_repo_return_repo_itself(
        repo=Repo(path=tmp_local_empty_git_repo)
    )


@skip_for_windows_py_lt_3_9
def test_clone_on_remote_repo_if_repo_is_remote_url_then_clone_and_set_repo_to_its_local_path():
    with patch("gto.git_utils.git_clone") as mocked_git_clone:
        mocked_git_clone.side_effect = git_clone
        local_repo = decorated_read_func(
            repo=tests.resources.SAMPLE_HTTP_REMOTE_REPO, spam=0, jam=3
        )
        mocked_git_clone.assert_called_once_with(
            repo=tests.resources.SAMPLE_HTTP_REMOTE_REPO, dir=local_repo
        )


@only_for_windows_py_lt_3_8
def test_if_repo_is_remote_url_and_windows_os_error_then_hint_win_with_py_lt_3_9_may_be_the_cause():
    with pytest.raises(OSError) as e:
        decorated_read_func(repo=tests.resources.SAMPLE_HTTP_REMOTE_REPO, spam=0, jam=3)
    assert e.type in (NotADirectoryError, PermissionError)
    assert "windows" in str(e)
    assert "python" in str(e)
    assert "< 3.9" in str(e)


@pytest.mark.parametrize(
    "repo",
    (
        "/local/path",
        "/local/path",
        ".",
    ),
)
def test_is_url_of_remote_repo_if_local_url_then_true(repo: str):
    assert is_url_of_remote_repo(repo=repo) is False


@pytest.mark.parametrize(
    "repo",
    (
        tests.resources.SAMPLE_HTTP_REMOTE_REPO,
        tests.resources.SAMPLE_HTTP_REMOTE_REPO_WITHOUT_DOT_GIT_SUFFIX,
    ),
)
def test_is_url_of_remote_repo_if_remote_url_then_true(repo: str):
    assert is_url_of_remote_repo(repo=repo) is True


@skip_for_windows_py_lt_3_9
@pytest.mark.parametrize(
    "repo",
    (
        tests.resources.SAMPLE_HTTP_REMOTE_REPO,
        tests.resources.SAMPLE_HTTP_REMOTE_REPO_WITHOUT_DOT_GIT_SUFFIX,
    ),
)
def test_git_clone_if_valid_remote_then_clone(repo: str):
    with TemporaryDirectory() as tmp_repo_dir:
        git_clone(repo=repo, dir=tmp_repo_dir)
        assert_dir_contain_git_repo(dir=tmp_repo_dir)


def test_git_push_tag_if_called_then_gitpython_corresponding_methods_are_correctly_invoked(
    with_mocked_repo_with_remote: tuple,
):
    (
        path,
        remote_name,
        MockedRepo,
        mocked_repo,
        mocked_remote,
    ) = with_mocked_repo_with_remote
    tag_name = "test_tag"

    git_push_tag(repo_path=path, tag_name=tag_name, remote_name=remote_name)

    MockedRepo.assert_called_once_with(path=path)
    mocked_repo.remote.assert_called_once_with(name=remote_name)
    mocked_remote.push.assert_called_once_with([tag_name])


def test_git_push_tag_if_called_with_delete_then_gitpython_corresponding_methods_are_correctly_invoked(
    with_mocked_repo_with_remote: tuple,
):
    (
        path,
        remote_name,
        MockedRepo,
        mocked_repo,
        mocked_remote,
    ) = with_mocked_repo_with_remote
    tag_name = "test_tag"

    git_push_tag(
        repo_path=path, tag_name=tag_name, delete=True, remote_name=remote_name
    )

    MockedRepo.assert_called_once_with(path=path)
    mocked_repo.remote.assert_called_once_with(name=remote_name)
    mocked_remote.push.assert_called_once_with(["--delete", tag_name])


def test_git_push_tag_if_error_then_exit_with_code_1(
    with_mocked_repo_with_remote: tuple,
):
    (
        path,
        remote_name,
        _,
        _,
        mocked_remote,
    ) = with_mocked_repo_with_remote
    mocked_remote.push.return_value.error = MagicMock()
    tag_name = "test_tag"

    with pytest.raises(GTOException) as error:
        git_push_tag(repo_path=path, tag_name=tag_name, remote_name=remote_name)

    assert f"git push {remote_name} {tag_name}" in error.value.msg
    assert (
        "Make sure your local repository is in sync with the remote" in error.value.msg
    )


def test_set_auto_push_on_remote_repo_if_not_remote_then_auto_push_is_not_changed(
    tmp_local_empty_git_repo,
):
    assert decorated_write_func(spam=37, repo=tmp_local_empty_git_repo, push=True)[0]
    assert not decorated_write_func(spam=37, repo=tmp_local_empty_git_repo, push=False)[
        0
    ]


@skip_for_windows_py_lt_3_9
def test_set_auto_push_on_remote_repo_if_remote_then_auto_push_is_set_to_true():
    assert decorated_write_func(
        spam=37, repo=tests.resources.SAMPLE_HTTP_REMOTE_REPO, push=True
    )[0]
    assert decorated_write_func(
        spam=37, repo=tests.resources.SAMPLE_HTTP_REMOTE_REPO, push=False
    )[0]


@skip_for_windows_py_lt_3_9
def test_git_add_and_commit_all_changes_if_files_not_changed_then_no_new_commit(
    tmp_local_git_repo_with_first_test_commit,
):
    git_add_and_commit_all_changes(
        repo_path=tmp_local_git_repo_with_first_test_commit[0],
        message=SECOND_TEST_COMMIT_MESSAGE,
    )
    assert_repo_as_expected_after_first_test_commit(
        repo_path=tmp_local_git_repo_with_first_test_commit[0]
    )


@skip_for_windows_py_lt_3_9
def test_git_add_and_commit_all_changes_if_tracked_file_is_changed_then_new_commit(
    tmp_local_git_repo_with_first_test_commit,
):
    with open(tmp_local_git_repo_with_first_test_commit[1], "a", encoding="utf") as f:
        f.write(SECOND_TEST_FILE_MODIFICATION)

    git_add_and_commit_all_changes(
        repo_path=tmp_local_git_repo_with_first_test_commit[0],
        message=SECOND_TEST_COMMIT_MESSAGE,
    )

    assert_repo_as_expected_after_second_test_commit(
        repo_path=tmp_local_git_repo_with_first_test_commit[0],
        second_commit_on_untracked_file=False,
    )


def test_git_commit_specific_files_if_untracked_file_is_changed_then_new_commit(
    tmp_local_git_repo_with_first_test_commit,
):
    untracked_file = (
        Path(tmp_local_git_repo_with_first_test_commit[0]) / TEST_COMMIT_UNTRACKED_FILE
    )
    with open(untracked_file, "w", encoding="utf") as f:
        f.write(SECOND_TEST_FILE_MODIFICATION)

    git_add_and_commit_all_changes(
        repo_path=tmp_local_git_repo_with_first_test_commit[0],
        message=SECOND_TEST_COMMIT_MESSAGE,
    )

    assert_repo_as_expected_after_second_test_commit(
        repo_path=tmp_local_git_repo_with_first_test_commit[0],
        second_commit_on_untracked_file=True,
    )


def test_stashed_changes_if_repo_has_no_ref_then_raise_exception(
    tmp_local_empty_git_repo,
):
    with pytest.raises(RuntimeError):
        with stashed_changes(repo_path=tmp_local_empty_git_repo):
            pass


@skip_for_windows_py_lt_3_9
def test_stashed_changes_if_tracked_file_was_changed_then_inside_with_statement_is_rolled_back(
    tmp_local_git_repo_with_first_test_commit,
):
    tracked_file, _ = change_tracked_file(
        repo_path=tmp_local_git_repo_with_first_test_commit[0]
    )

    with stashed_changes(repo_path=tmp_local_git_repo_with_first_test_commit[0]):
        with open(tracked_file, "r", encoding="utf") as f:
            assert f.read() == FIRST_TEST_FILE_MODIFICATION


@skip_for_windows_py_lt_3_9
def test_stashed_changes_if_tracked_file_was_changed_then_outside_with_statement_is_as_before(
    tmp_local_git_repo_with_first_test_commit,
):
    tracked_file, new_file_content = change_tracked_file(
        repo_path=tmp_local_git_repo_with_first_test_commit[0]
    )

    with stashed_changes(repo_path=tmp_local_git_repo_with_first_test_commit[0]):
        pass

    with open(tracked_file, "r", encoding="utf") as f:
        assert f.read() == new_file_content


@skip_for_windows_py_lt_3_9
def test_stashed_changes_if_tracked_file_was_changed_then_return_its_path(
    tmp_local_git_repo_with_first_test_commit,
):
    repo_path = tmp_local_git_repo_with_first_test_commit[0]
    tracked_file, _ = change_tracked_file(repo_path=repo_path)

    with stashed_changes(repo_path=repo_path) as (
        tracked,
        untracked,
    ):
        assert tracked == [tracked_file.relative_to(repo_path).as_posix()]
        assert len(untracked) == 0


@skip_for_windows_py_lt_3_9
def test_stashed_changes_if_untracked_file_was_changed_but_include_untracked_is_false_then_do_not_roll_back(
    tmp_local_git_repo_with_first_test_commit,
):
    untracked_file, _ = change_untracked_file(
        repo_path=tmp_local_git_repo_with_first_test_commit[0]
    )

    with stashed_changes(
        repo_path=tmp_local_git_repo_with_first_test_commit[0], include_untracked=False
    ):
        assert untracked_file.is_file()


@skip_for_windows_py_lt_3_9
def test_stashed_changes_if_untracked_file_was_changed_then_inside_with_statement_is_rolled_back(
    tmp_local_git_repo_with_first_test_commit,
):
    untracked_file, _ = change_untracked_file(
        repo_path=tmp_local_git_repo_with_first_test_commit[0]
    )

    with stashed_changes(
        repo_path=tmp_local_git_repo_with_first_test_commit[0], include_untracked=True
    ):
        assert not untracked_file.is_file()


@skip_for_windows_py_lt_3_9
def test_stashed_changes_if_untracked_file_was_changed_then_outside_with_statement_is_as_before(
    tmp_local_git_repo_with_first_test_commit,
):
    untracked_file, new_file_content = change_untracked_file(
        repo_path=tmp_local_git_repo_with_first_test_commit[0]
    )

    with stashed_changes(
        repo_path=tmp_local_git_repo_with_first_test_commit[0], include_untracked=True
    ):
        pass

    with open(untracked_file, "r", encoding="utf") as f:
        assert f.read() == new_file_content


@skip_for_windows_py_lt_3_9
def test_stashed_changes_if_untracked_file_was_changed_then_return_its_path(
    tmp_local_git_repo_with_first_test_commit,
):
    repo_path = tmp_local_git_repo_with_first_test_commit[0]
    untracked_file, _ = change_untracked_file(repo_path=repo_path)

    with stashed_changes(repo_path=repo_path, include_untracked=True) as (
        tracked,
        untracked,
    ):
        assert len(tracked) == 0
        assert untracked == [untracked_file.relative_to(repo_path).as_posix()]


def test_commit_produced_changes_on_auto_commit_if_no_auto_commit_argument_then_don_t_stash_or_commit(
    mocked_f_decorated_with_commit_produced_changes_on_auto_commit,
):
    (
        f,
        f_spy,
        mocked_stashed_changes,
        mocked_git_add_and_commit_all_changes,
        _,
        _,
    ) = mocked_f_decorated_with_commit_produced_changes_on_auto_commit

    result = f()

    mocked_stashed_changes.assert_not_called()
    mocked_git_add_and_commit_all_changes.assert_not_called()
    f_spy.assert_called_once_with()
    assert result == f_spy.return_value


def test_commit_produced_changes_on_auto_commit_if_auto_commit_is_false_then_don_t_stash_or_commit(
    mocked_f_decorated_with_commit_produced_changes_on_auto_commit,
):
    (
        f,
        f_spy,
        mocked_stashed_changes,
        mocked_git_add_and_commit_all_changes,
        _,
        _,
    ) = mocked_f_decorated_with_commit_produced_changes_on_auto_commit

    result = f(commit=False)

    mocked_stashed_changes.assert_not_called()
    mocked_git_add_and_commit_all_changes.assert_not_called()
    f_spy.assert_called_once_with(commit=False)
    assert result == f_spy.return_value


def test_commit_produced_changes_on_auto_commit_if_auto_commit_but_no_repo_then_exception_and_don_t_stash_or_commit(
    mocked_f_decorated_with_commit_produced_changes_on_auto_commit,
):
    (
        f,
        f_spy,
        mocked_stashed_changes,
        mocked_git_add_and_commit_all_changes,
        _,
        _,
    ) = mocked_f_decorated_with_commit_produced_changes_on_auto_commit

    with pytest.raises(ValueError):
        f(commit=True)

    mocked_stashed_changes.assert_not_called()
    mocked_git_add_and_commit_all_changes.assert_not_called()
    f_spy.assert_not_called()


def test_commit_produced_changes_on_auto_commit_if_auto_commit_is_true_then_stash_and_commit_in_repo(
    mocked_f_decorated_with_commit_produced_changes_on_auto_commit,
):
    (
        f,
        f_spy,
        _,
        _,
        mock_manager,
        generate_test_commit_message,
    ) = mocked_f_decorated_with_commit_produced_changes_on_auto_commit
    repo_path = "a/repo/path"

    result = f(repo=repo_path, commit=True)

    expected_calls = [
        call.stashed_changes(repo_path=repo_path, include_untracked=True),
        call.stashed_changes().__enter__(),
        call.spy(repo=repo_path, commit=True),
        call.git_add_and_commit_all_changes(
            repo_path=repo_path, message=generate_test_commit_message(commit=True)
        ),
        call.stashed_changes().__exit__(None, None, None),
    ]

    assert mock_manager.mock_calls == expected_calls
    assert result == f_spy.return_value


@skip_for_windows_py_lt_3_9
def test_commit_produced_changes_on_auto_commit_if_f_changes_tracked_file_already_dangling_then_raise_exception_and_roll_back(
    tmp_local_git_repo_with_first_test_commit,
):
    repo, tracked_file = tmp_local_git_repo_with_first_test_commit

    with open(tracked_file, "a", encoding="utf") as f:
        f.write(f" - {SECOND_TEST_FILE_MODIFICATION}")

    @commit_produced_changes_on_commit()
    def f_to_test(repo: str, commit: bool):  # pylint: disable=unused-argument
        with open(tracked_file, "a", encoding="utf") as f:
            f.write(" - change made by f")

    with pytest.raises(GTOException):
        f_to_test(repo=repo, commit=True)

    with open(tracked_file, "r", encoding="utf") as f:
        assert (
            f.read()
            == f"{FIRST_TEST_FILE_MODIFICATION} - {SECOND_TEST_FILE_MODIFICATION}"
        )


@skip_for_windows_py_lt_3_9
def test_commit_produced_changes_on_auto_commit_if_f_changes_untracked_file_already_dangling_then_raise_exception_and_roll_back(
    tmp_local_git_repo_with_first_test_commit,
):
    repo, _ = tmp_local_git_repo_with_first_test_commit
    untracked_file = (Path(repo) / TEST_COMMIT_UNTRACKED_FILE).as_posix()

    with open(untracked_file, "w", encoding="utf") as f:
        f.write(f"{FIRST_TEST_FILE_MODIFICATION}")

    @commit_produced_changes_on_commit()
    def f_to_test(repo: str, commit: bool):  # pylint: disable=unused-argument
        with open(untracked_file, "a", encoding="utf") as f:
            f.write(" - change made by f")

    with pytest.raises(GTOException):
        f_to_test(repo=repo, commit=True)

    with open(untracked_file, "r", encoding="utf") as f:
        assert f.read() == f"{FIRST_TEST_FILE_MODIFICATION}"


@skip_for_windows_py_lt_3_9
def test_git_push_if_called_then_corresponding_gitpython_functions_are_called(
    tmp_local_empty_git_repo,
):
    with patch("gto.git_utils.git.Repo") as MockedRepo:
        git_push(repo_path=tmp_local_empty_git_repo)

    MockedRepo.assert_called_once_with(path=tmp_local_empty_git_repo)
    MockedRepo.return_value.git.push.assert_called_once_with()


def test_push_on_auto_push_if_auto_push_and_repo_not_provided_then_raise_exception(
    mocked_f_decorated_with_push_on_auto_push,
):
    f, _, _, _, _ = mocked_f_decorated_with_push_on_auto_push

    with pytest.raises(ValueError):
        f(push=True)


def test_push_on_auto_push_if_auto_push_then_set_auto_commit_to_true(
    mocked_f_decorated_with_push_on_auto_push,
):
    f, f_spy, repo_path, _, _ = mocked_f_decorated_with_push_on_auto_push

    result = f(repo=repo_path, commit=False, push=True)

    f_spy.assert_called_once_with(repo=repo_path, commit=True, push=True)
    assert result == f_spy.return_value


def test_push_on_auto_push_if_auto_push_false_then_git_push_is_not_called(
    mocked_f_decorated_with_push_on_auto_push,
):
    f, f_spy, _, _, mock_manager = mocked_f_decorated_with_push_on_auto_push

    result = f(push=False)

    expected_calls = [
        call.spy(push=False),
    ]
    assert mock_manager.mock_calls == expected_calls
    assert result == f_spy.return_value


def test_push_on_auto_push_if_auto_push_true_then_git_push_is_called_after_f(
    mocked_f_decorated_with_push_on_auto_push,
):
    f, f_spy, repo_path, _, mock_manager = mocked_f_decorated_with_push_on_auto_push

    result = f(repo=repo_path, push=True)

    expected_calls = [
        call.spy(repo=repo_path, commit=True, push=True),
        call.git_push(
            repo_path=repo_path,
        ),
    ]
    assert mock_manager.mock_calls == expected_calls
    assert result == f_spy.return_value


def test_push_on_auto_push_if_git_pull_fails_then_raise_gto_exception(
    mocked_f_decorated_with_push_on_auto_push,
):
    (
        f,
        _,
        repo_path,
        mocked_git_push,
        _,
    ) = mocked_f_decorated_with_push_on_auto_push
    git_push_error_message = "This was the mistake..."
    mocked_git_push.side_effect = Exception(git_push_error_message)

    with pytest.raises(GTOException) as e:
        f(repo=repo_path, push=True)

    assert "It was not possible to run `git push`" in e.value.msg
    assert git_push_error_message in e.value.msg


@set_push_on_remote_repo
def decorated_write_func(
    spam: int, repo: Union[Repo, str], push: bool
):  # pylint: disable=unused-argument
    return push, repo


@clone_on_remote_repo
def decorated_read_func(
    spam: int, repo: Union[Repo, str], jam: int
):  # pylint: disable=unused-argument
    return repo


def assert_f_called_with_repo_return_repo_itself(repo: Union[str, Repo]) -> None:
    assert decorated_read_func(0, repo, 3) is repo
    assert decorated_read_func(0, repo, jam=3) is repo
    assert decorated_read_func(0, jam=3, repo=repo) is repo
    assert decorated_read_func(spam=0, jam=3, repo=repo) is repo


@pytest.fixture
def tmp_local_empty_git_repo() -> str:
    tmp_repo_dir = TemporaryDirectory()  # pylint: disable=consider-using-with
    Repo.init(path=tmp_repo_dir.name)
    yield tmp_repo_dir.name
    tmp_repo_dir.cleanup()


@pytest.fixture
def tmp_local_git_repo_with_first_test_commit(tmp_local_empty_git_repo) -> str:
    new_file_path = Path(tmp_local_empty_git_repo) / TEST_COMMIT_FILE
    repo = Repo(path=tmp_local_empty_git_repo)
    with open(new_file_path, "w", encoding="utf") as test_file:
        test_file.write(FIRST_TEST_FILE_MODIFICATION)
    with open(
        Path(tmp_local_empty_git_repo) / "README.md", "w", encoding="utf"
    ) as readme:
        readme.write("Read me")
    repo.index.add(items=[test_file.name, readme.name])
    repo.index.commit(message=FIRST_TEST_COMMIT_MESSAGE)
    yield tmp_local_empty_git_repo, new_file_path.as_posix()


@pytest.fixture
def mocked_f_decorated_with_commit_produced_changes_on_auto_commit() -> Tuple[
    Callable, MagicMock, MagicMock, MagicMock, MagicMock, Callable
]:
    f_spy = MagicMock()
    f_spy.return_value = MagicMock()

    def generate_test_commit_message(commit: bool) -> str:
        return f"commit message with argument {commit}"

    @commit_produced_changes_on_commit(message_generator=generate_test_commit_message)
    def f(*args, **kwargs):
        return f_spy(*args, **kwargs)

    with patch("gto.git_utils.stashed_changes") as mocked_stashed_changes:
        mocked_stashed_changes.return_value.__enter__.return_value = [], []
        with patch(
            "gto.git_utils._get_repo_changed_tracked_and_untracked_files"
        ) as mocked_get_repo_changed_tracked_and_untracked_files:
            mocked_get_repo_changed_tracked_and_untracked_files.return_value = [
                TEST_COMMIT_FILE
            ], []
            with patch(
                "gto.git_utils.git_add_and_commit_all_changes"
            ) as mocked_git_add_and_commit_all_changes:
                mock_manager = MagicMock()
                mock_manager.attach_mock(mocked_stashed_changes, "stashed_changes")
                mock_manager.attach_mock(f_spy, "spy")
                mock_manager.attach_mock(
                    mocked_git_add_and_commit_all_changes,
                    "git_add_and_commit_all_changes",
                )

                yield f, f_spy, mocked_stashed_changes, mocked_git_add_and_commit_all_changes, mock_manager, generate_test_commit_message


@pytest.fixture
def mocked_f_decorated_with_push_on_auto_push() -> Tuple[
    Callable, MagicMock, str, MagicMock, MagicMock
]:
    repo_path = "my/repo"
    f_spy = MagicMock()

    @push_on_push
    def f(*args, **kwargs):
        return f_spy(*args, **kwargs)

    with patch("gto.git_utils.git_push") as mocked_git_push:
        mock_manager = MagicMock()
        mock_manager.attach_mock(f_spy, "spy")
        mock_manager.attach_mock(mocked_git_push, "git_push")

        yield f, f_spy, repo_path, mocked_git_push, mock_manager


def change_tracked_file(repo_path: str) -> Tuple[Path, str]:
    tracked_file = Path(repo_path) / TEST_COMMIT_FILE
    new_file_content = "Same changes"
    with open(tracked_file, "w", encoding="utf") as f:
        f.write(new_file_content)
    return tracked_file, new_file_content


def change_untracked_file(
    repo_path: str,
) -> Tuple[Path, str]:
    untracked_file = Path(repo_path) / TEST_COMMIT_UNTRACKED_FILE
    new_file_content = "Same changes to an untracked file"
    with open(untracked_file, "w", encoding="utf") as f:
        f.write(new_file_content)
    return untracked_file, new_file_content


def assert_dir_contain_git_repo(dir: str) -> None:
    assert (Path(dir) / ".git").is_dir()
    assert (Path(dir) / ".git/HEAD").is_file()


@pytest.fixture
def with_mocked_repo_with_remote() -> tuple:
    mocked_remote = MagicMock()
    mocked_push_info = MagicMock()
    mocked_push_info.error = None
    mocked_remote.push.return_value = mocked_push_info
    mocked_repo = MagicMock()
    path = "git_repo_path"
    remote_name = "git_remote_name"

    with patch("gto.git_utils.git.Repo") as MockedRepo:
        MockedRepo.return_value = mocked_repo
        mocked_repo.remote.return_value = mocked_remote
        yield path, remote_name, MockedRepo, mocked_repo, mocked_remote


def assert_repo_without_commit(repo_path: str) -> None:
    with pytest.raises(ValueError) as e:
        Repo(path=repo_path).iter_commits()
    assert "Reference at 'refs/heads/master' does not exist" in str(e)


def assert_repo_as_expected_after_first_test_commit(repo_path: str) -> None:
    assert (
        list(Repo(path=repo_path).iter_commits())[0].message
        == FIRST_TEST_COMMIT_MESSAGE
    )
    assert_test_commit_file_content(
        repo_path=repo_path,
        file_name=TEST_COMMIT_FILE,
        expected_content=FIRST_TEST_FILE_MODIFICATION,
    )


def assert_repo_as_expected_after_second_test_commit(
    repo_path: str, second_commit_on_untracked_file: bool
) -> None:
    assert [c.message for c in Repo(path=repo_path).iter_commits()] == [
        SECOND_TEST_COMMIT_MESSAGE,
        FIRST_TEST_COMMIT_MESSAGE,
    ]
    if second_commit_on_untracked_file:
        assert_test_commit_file_content(
            repo_path=repo_path,
            file_name=TEST_COMMIT_FILE,
            expected_content=FIRST_TEST_FILE_MODIFICATION,
        )
        assert_test_commit_file_content(
            repo_path=repo_path,
            file_name=TEST_COMMIT_UNTRACKED_FILE,
            expected_content=SECOND_TEST_FILE_MODIFICATION,
        )
    else:
        assert_test_commit_file_content(
            repo_path=repo_path,
            file_name=TEST_COMMIT_FILE,
            expected_content=FIRST_TEST_FILE_MODIFICATION
            + SECOND_TEST_FILE_MODIFICATION,
        )


def assert_test_commit_file_content(
    repo_path: str, file_name: str, expected_content: str
) -> None:
    with TemporaryDirectory() as td:
        Repo.clone_from(url=repo_path, to_path=td)
        with open(Path(td) / file_name, "r", encoding="utf") as f:
            assert f.read() == expected_content


TEST_COMMIT_FILE = "foo.txt"
TEST_COMMIT_UNTRACKED_FILE = "untracked_foo.txt"
FIRST_TEST_FILE_MODIFICATION = "First file modification"
SECOND_TEST_FILE_MODIFICATION = "Second file modification"
FIRST_TEST_COMMIT_MESSAGE = "first test commit"
SECOND_TEST_COMMIT_MESSAGE = "second test commit"



from typing import Optional


def generate_annotate_commit_message(
    name: str, type: Optional[str] = None, path: Optional[str] = None
) -> str:
    return (
        f"Annotate artifact `{name}`"
        f"{f' of type `{type}`' if type is not None else ''}"
        f"{f' with path `{path}`' if path is not None else ''}"
    )


def generate_remove_commit_message(name: str) -> str:
    return f"Remove annotation for artifact {name}"


def generate_empty_commit_message() -> str:
    return ""



import re
from enum import Enum

COMMIT = "commit"
REF = "ref"
TAG = "tag"

ARTIFACT = "artifact"
ACTION = "action"
TYPE = "type"
NAME = "name"
PATH = "path"
VERSION = "version"
STAGE = "stage"
COUNTER = "counter"


class Action(Enum):
    CREATE = "create"
    DEPRECATE = "deprecate"
    DEREGISTER = "deregister"
    REGISTER = "register"
    ASSIGN = "assign"
    UNASSIGN = "unassign"


name = "[a-z][a-z0-9-/]*[a-z0-9]"
semver = r"(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"
counter = "?P<counter>[0-9]+"
name_regexp = re.compile(f"^{name}$")
tag_regexp = re.compile(
    f"^(?P<artifact>{name})(((#(?P<stage>{name})|@(?P<version>v{semver}))(?P<cancel>!?))|@((?P<deprecated>deprecated)|(?P<created>created)))(#({counter}))?$"
)
shortcut_regexp = re.compile(
    f"^(?P<artifact>{name})(((#(?P<stage>{name})|@(?P<latest>latest)|@(?P<greatest>greatest))))$"
)

# taken from https://stackoverflow.com/a/22312124/19782654, modified to include url without .git at the end
remote_git_repo_regex = re.compile(
    r"((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(/)?"
)


def mark_artifact_unregistered(artifact_name):
    return f"*{artifact_name}"


class VersionSort(Enum):
    SemVer = "semver"
    Timestamp = "timestamp"


ASSIGNMENTS_PER_VERSION = -1
VERSIONS_PER_STAGE = 1



import re
from datetime import datetime
from enum import Enum
from typing import FrozenSet, Iterable, Optional, Union

import git
from pydantic import BaseModel

from .base import (
    Artifact,
    Assignment,
    BaseManager,
    BaseRegistryState,
    Deprecation,
    Deregistration,
    Registration,
    Unassignment,
)
from .constants import (
    ACTION,
    COUNTER,
    NAME,
    STAGE,
    TAG,
    VERSION,
    Action,
    tag_regexp,
)
from .exceptions import (
    InvalidTagName,
    MissingArg,
    NotImplementedInGTO,
    RefNotFound,
    TagExists,
    TagNotFound,
    UnknownAction,
)

COUNT_DELIMITER = "#"

TagTemplates = {
    # Action.CREATE: "{artifact}@",
    Action.DEPRECATE: "{artifact}@deprecated",
    Action.REGISTER: "{artifact}@{version}",
    Action.DEREGISTER: "{artifact}@{version}!",
    Action.ASSIGN: "{artifact}#{stage}",
    Action.UNASSIGN: "{artifact}#{stage}!",
}


def name_tag(
    action: Action,
    artifact: str,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    repo: Optional[git.Repo] = None,
    simple: bool = False,
):
    if action not in TagTemplates:
        raise UnknownAction(action=action)

    tag = TagTemplates[action].format(artifact=artifact, version=version, stage=stage)
    if simple:
        return tag
    if repo is None:
        raise MissingArg(arg="repo")
    counter = 0
    for t in repo.tags:
        parsed = parse_name(t.name, raise_on_fail=False)  # type: ignore
        if (
            parsed
            and parsed[NAME] == artifact
            and COUNTER in parsed
            and parsed[COUNTER] > counter
        ):
            counter = parsed[COUNTER]
    return f"{tag}{COUNT_DELIMITER}{counter+1}"


def parse_name(name: str, raise_on_fail: bool = True):

    match = re.search(tag_regexp, name)
    if raise_on_fail and not match:
        raise InvalidTagName(name)
    if match:
        parsed = {NAME: match["artifact"]}
        if match["deprecated"]:
            parsed[ACTION] = Action.DEPRECATE
        if match[VERSION]:
            parsed[VERSION] = match[VERSION]
            parsed[ACTION] = (
                Action.DEREGISTER if match["cancel"] == "!" else Action.REGISTER
            )
        if match[STAGE]:
            parsed[STAGE] = match[STAGE]
            parsed[ACTION] = (
                Action.UNASSIGN if match["cancel"] == "!" else Action.ASSIGN
            )
        if match[COUNTER]:
            parsed[COUNTER] = int(match[COUNTER])
        return parsed
    return {}


class NAME_REFERENCE(Enum):
    TAG = TAG
    NAME = NAME


def parse_name_reference(name: str):
    parsed = parse_name(name, raise_on_fail=False)
    if not parsed:
        return NAME_REFERENCE.NAME, name
    return NAME_REFERENCE.TAG, parsed


class Tag(BaseModel):
    action: Action
    name: str
    version: Optional[str]
    stage: Optional[str]
    created_at: datetime
    tag: git.Tag

    class Config:
        arbitrary_types_allowed = True


def parse_tag(tag: git.Tag):
    return Tag(
        tag=tag,
        created_at=datetime.fromtimestamp(tag.tag.tagged_date),
        **parse_name(tag.name),
    )


def find(
    action: Union[Action, FrozenSet[Action]] = None,
    name: Optional[str] = None,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    repo: Optional[git.Repo] = None,
    sort: str = "by_time",
    tags: Optional[Iterable[git.Tag]] = None,
):
    if isinstance(action, Action):
        action = frozenset([action])
    if tags is None:
        if repo is None:
            raise MissingArg(arg="repo")
        tags = [t for t in repo.tags if parse_name(t.name, raise_on_fail=False)]
    if action:
        tags = [t for t in tags if parse_name(t.name)[ACTION] in action]
    if name:
        tags = [t for t in tags if parse_name(t.name).get(NAME) == name]
    if version:
        tags = [t for t in tags if parse_name(t.name).get(VERSION) == version]
    if stage:
        tags = [t for t in tags if parse_name(t.name).get(STAGE) == stage]
    # remove lightweight tags - better to do later so the function is faster
    tags = [t for t in tags if t.tag is not None]
    if sort == "by_time":
        tags = sorted(tags, key=lambda t: t.tag.tagged_date)
    else:
        raise NotImplementedError(f"Unknown sort: {sort}")
    return tags


def create_tag(
    repo: git.Repo,
    name: str,
    ref: str,
    message: str,
    tagger: str = None,
    tagger_email: str = None,
):
    try:
        repo.commit(ref)
    except (ValueError, git.BadName) as e:
        raise RefNotFound(ref=ref) from e
    if name in repo.refs:
        raise TagExists(name=name)

    env = {}
    if tagger:
        env["GIT_COMMITTER_NAME"] = tagger
    if tagger_email:
        env["GIT_COMMITTER_EMAIL"] = tagger_email

    repo.git.tag(["-a", name, "-m", message, ref], env=env)


def delete_tag(repo: git.Repo, name: str):
    try:
        repo.delete_tag(name)
    except git.BadName as e:
        raise TagNotFound(name=name) from e


def index_tag(artifact: Artifact, tag: git.TagReference) -> Artifact:
    event: Union[Deprecation, Registration, Deregistration, Assignment, Unassignment]
    mtag = parse_tag(tag)
    if mtag.action == Action.REGISTER:
        event = Registration(
            artifact=mtag.name,
            version=mtag.version,
            created_at=mtag.created_at,
            author=tag.tag.tagger.name,
            author_email=tag.tag.tagger.email,
            message=tag.tag.message,
            commit_hexsha=tag.commit.hexsha,
            tag=tag.name,
        )
    elif mtag.action == Action.DEREGISTER:
        event = Deregistration(
            artifact=mtag.name,
            version=mtag.version,
            created_at=mtag.created_at,
            author=tag.tag.tagger.name,
            author_email=tag.tag.tagger.email,
            message=tag.tag.message,
            commit_hexsha=tag.commit.hexsha,
            tag=tag.name,
        )
    elif mtag.action in (Action.ASSIGN, Action.UNASSIGN):
        version = artifact.find_version(
            commit_hexsha=tag.commit.hexsha, create_new=True
        ).version  # type: ignore
        if mtag.action == Action.ASSIGN:
            event = Assignment(
                artifact=mtag.name,
                version=version,
                stage=mtag.stage,
                created_at=mtag.created_at,
                author=tag.tag.tagger.name,
                author_email=tag.tag.tagger.email,
                message=tag.tag.message,
                commit_hexsha=tag.commit.hexsha,
                tag=tag.name,
            )
        else:
            event = Unassignment(
                artifact=mtag.name,
                version=version,
                stage=mtag.stage,
                created_at=mtag.created_at,
                author=tag.tag.tagger.name,
                author_email=tag.tag.tagger.email,
                message=tag.tag.message,
                commit_hexsha=tag.commit.hexsha,
                tag=tag.name,
            )
    elif mtag.action == Action.DEPRECATE:
        event = Deprecation(
            artifact=mtag.name,
            created_at=mtag.created_at,
            author=tag.tag.tagger.name,
            author_email=tag.tag.tagger.email,
            message=tag.tag.message,
            commit_hexsha=tag.commit.hexsha,
            tag=tag.name,
        )
    artifact.add_event(event)
    return artifact


class TagManager(BaseManager):  # pylint: disable=abstract-method
    def update_state(self, state: BaseRegistryState) -> BaseRegistryState:
        # tags are sorted and then indexed by timestamp
        # this is important to check that history is not broken
        for tag in find(repo=self.repo, action=self.actions):
            state.update_artifact(
                index_tag(
                    state.find_artifact(
                        parse_name(tag.tag.tag)["name"], create_new=True
                    ),
                    tag,
                )
            )
        return state


class TagArtifactManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.CREATE, Action.DEPRECATE))

    def create(self):  # pylint: disable=no-self-use
        raise NotImplementedInGTO(
            "If you want to create artifact, register a version or assign a stage for it"
        )

    def deprecate(
        self,
        name,
        ref,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.DEPRECATE,
            name,
            repo=self.repo,
            simple=simple,
        )
        create_tag(
            self.repo,
            tag,
            ref=ref,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag


class TagVersionManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.REGISTER, Action.DEREGISTER))

    def register(
        self,
        name,
        version,
        ref,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.REGISTER, name, version=version, repo=self.repo, simple=simple
        )
        create_tag(
            self.repo,
            tag,
            ref=ref,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag

    def deregister(
        self,
        name,
        version,
        ref,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.DEREGISTER, name, version=version, repo=self.repo, simple=simple
        )
        create_tag(
            self.repo,
            tag,
            ref=ref,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag


class TagStageManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.ASSIGN, Action.UNASSIGN))

    def assign(
        self,
        name,
        stage,
        ref,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> str:
        tag = name_tag(Action.ASSIGN, name, stage=stage, repo=self.repo, simple=simple)
        create_tag(
            self.repo,
            tag,
            ref=ref,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag

    def unassign(
        self,
        name,
        stage,
        ref,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> str:
        tag = name_tag(
            Action.UNASSIGN, name, stage=stage, repo=self.repo, simple=simple
        )
        create_tag(
            self.repo,
            tag,
            ref=ref,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag



import inspect
import logging
from contextlib import contextmanager
from functools import wraps
from tempfile import TemporaryDirectory
from typing import Callable, Dict, List, Tuple

import git

from gto.commit_message_generator import generate_empty_commit_message
from gto.constants import remote_git_repo_regex
from gto.exceptions import GTOException, WrongArgs


def clone_on_remote_repo(f: Callable):
    @wraps(f)
    def wrapped_f(*args, **kwargs):
        kwargs = _turn_args_into_kwargs(f, args, kwargs)

        if isinstance(kwargs["repo"], str) and is_url_of_remote_repo(
            repo=kwargs["repo"]
        ):
            try:
                with cloned_git_repo(repo=kwargs["repo"]) as tmp_dir:
                    kwargs["repo"] = tmp_dir
                    return f(**kwargs)
            except (NotADirectoryError, PermissionError) as e:
                raise e.__class__(
                    "Are you using windows with python < 3.9? "
                    "This may be the reason of this error: https://bugs.python.org/issue42796. "
                    "Consider upgrading python."
                ) from e

        return f(**kwargs)

    return wrapped_f


def set_push_on_remote_repo(f: Callable):
    @wraps(f)
    def wrapped_f(*args, **kwargs):
        kwargs = _turn_args_into_kwargs(f, args, kwargs)

        if isinstance(kwargs["repo"], str) and is_url_of_remote_repo(
            repo=kwargs["repo"]
        ):
            kwargs["push"] = True

        return f(**kwargs)

    return wrapped_f


def commit_produced_changes_on_commit(
    message_generator: Callable[..., str] = generate_empty_commit_message
):
    """
    The function `message_generator` can use any argument that the decorated function has.

    Example: here we are using the argument b of the function f to generate the commit message

        def create_message(b: str) -> str:
            return "commit message with b={b}"

        @commit_produced_changes_on_commit(message_generator=create_message)
        def f(a: str, b: str, c: str):
            ...

    """

    def generate_commit_message(**kwargs) -> str:
        kwargs_for_message_generator = {
            k: kwargs.get(k, None)
            for k in inspect.getfullargspec(message_generator).args
        }
        return message_generator(**kwargs_for_message_generator)

    def wrap(f: Callable):
        @wraps(f)
        def wrapped_f(*args, **kwargs):
            kwargs = _turn_args_into_kwargs(f, args, kwargs)

            if kwargs.get("commit", False) is True:
                if "repo" in kwargs:
                    with stashed_changes(
                        repo_path=kwargs["repo"], include_untracked=True
                    ) as (stashed_tracked, stashed_untracked):
                        result = f(**kwargs)
                        if are_files_in_repo_changed(
                            repo_path=kwargs["repo"],
                            files=stashed_tracked + stashed_untracked,
                        ):
                            _reset_repo_to_head(repo_path=kwargs["repo"])
                            raise GTOException(
                                msg="The command would have changed files that were not committed, "
                                "automated committing is not possible.\n"
                                "Suggested action: Commit the changes and re-run this command."
                            )
                        git_add_and_commit_all_changes(
                            repo_path=kwargs["repo"],
                            message=generate_commit_message(**kwargs),
                        )
                else:
                    raise ValueError(
                        "Function decorated with commit_produced_changes_on_commit was called with "
                        "`commit=True` but `repo` was not provided."
                        "Argument `repo` is necessary."
                    )
            else:
                result = f(**kwargs)

            return result

        return wrapped_f

    return wrap


def push_on_push(f: Callable):
    @wraps(f)
    def wrapped_f(*args, **kwargs):
        kwargs = _turn_args_into_kwargs(f, args, kwargs)
        if kwargs.get("push", False) is True:
            kwargs["commit"] = True
            result = f(**kwargs)
            if "repo" in kwargs:
                try:
                    git_push(repo_path=kwargs["repo"])
                except Exception as e:
                    raise GTOException(  # pylint: disable=raise-missing-from
                        "It was not possible to run `git push`. "
                        "The detailed error message was:\n"
                        f"{str(e)}"
                    )
            else:
                raise ValueError(
                    "Function decorated with push_on_push was called with "
                    "`push=True` but `repo` was not provided."
                    "Argument `repo` is necessary."
                )
        else:
            result = f(**kwargs)
        return result

    return wrapped_f


def are_files_in_repo_changed(repo_path: str, files: List[str]) -> bool:
    tracked, untracked = _get_repo_changed_tracked_and_untracked_files(
        repo_path=repo_path
    )
    return (
        len(set(files).intersection(tracked)) > 0
        or len(set(files).intersection(untracked)) > 0
    )


def is_url_of_remote_repo(repo: str) -> bool:
    if remote_git_repo_regex.fullmatch(repo) is not None:
        logging.debug("%s recognized as remote git repo", repo)
        return True

    logging.debug("%s NOT recognized as remote git repo", repo)
    return False


@contextmanager
def cloned_git_repo(repo: str):
    tmp_dir = TemporaryDirectory()
    logging.debug("create temporary directory %s", tmp_dir)
    git_clone(repo=repo, dir=tmp_dir.name)
    yield tmp_dir.name
    logging.debug("delete temporary directory %s", tmp_dir)
    tmp_dir.cleanup()


def git_clone(repo: str, dir: str) -> None:
    logging.debug("clone %s in directory %s", repo, dir)
    git.Repo.clone_from(url=repo, to_path=dir)


def git_push_tag(
    repo_path: str, tag_name: str, delete: bool = False, remote_name: str = "origin"
) -> None:
    repo = git.Repo(path=repo_path)
    remote = repo.remote(name=remote_name)
    if not hasattr(remote, "url"):
        raise WrongArgs(
            f"provided repo_path={repo_path} does not appear to have a remote to push to"
        )
    logging.debug(
        "push %s tag %s from directory %s to remote %s with url %s",
        "--delete" if delete else "",
        tag_name,
        repo_path,
        remote_name,
        remote.url,
    )
    remote_push_args = [tag_name]
    if delete:
        remote_push_args = ["--delete"] + remote_push_args
    push_info = remote.push(remote_push_args)
    if push_info.error is not None:
        raise GTOException(
            msg=f"The command `git push {remote_name} {' '.join(remote_push_args)}` failed. "
            f"Make sure your local repository is in sync with the remote."
        )


def git_push(repo_path: str) -> None:
    git.Repo(path=repo_path).git.push()


def git_add_and_commit_all_changes(repo_path: str, message: str) -> None:
    repo = git.Repo(path=repo_path)
    tracked, untracked = _get_repo_changed_tracked_and_untracked_files(
        repo_path=repo_path
    )
    if len(tracked) + len(untracked) > 0:
        logging.debug("Adding to the index the untracked files %s", untracked)
        logging.debug("Add and commit changes to files %s", tracked + untracked)
        repo.index.add(items=tracked + untracked)
        repo.index.commit(message=message)


@contextmanager
def stashed_changes(repo_path: str, include_untracked: bool = False):
    repo = git.Repo(path=repo_path)
    if len(repo.refs) == 0:
        raise RuntimeError(
            "Cannot stash because repository has no ref. Please create a first commit."
        )

    tracked, untracked = _get_repo_changed_tracked_and_untracked_files(
        repo_path=repo_path
    )

    stash_arguments = ["push"]
    if include_untracked:
        stash_arguments += ["--include-untracked"]
    else:
        untracked = []

    try:
        if len(tracked + untracked) > 0:
            repo.git.stash(stash_arguments)
        yield tracked, untracked
    finally:
        if len(tracked + untracked) > 0:
            repo.git.stash("pop")


def _reset_repo_to_head(repo_path: str) -> None:
    repo = git.Repo(path=repo_path)
    repo.git.stash(["push", "--include-untracked"])
    repo.git.stash(["drop"])


def _get_repo_changed_tracked_and_untracked_files(
    repo_path: str,
) -> Tuple[List[str], List[str]]:
    repo = git.Repo(path=repo_path)
    return [item.a_path for item in repo.index.diff(None)], repo.untracked_files


def _turn_args_into_kwargs(
    f: Callable, args: tuple, kwargs: Dict[str, object]
) -> Dict[str, object]:
    kwargs_complement = {
        k: args[i]
        for i, k in enumerate(inspect.signature(f).parameters.keys())
        if i < len(args)
    }
    kwargs.update(kwargs_complement)
    return kwargs
